from __future__ import unicode_literals

from celery import chain

from waldur_core.core import executors as core_executors
from waldur_core.core import tasks as core_tasks
from waldur_core.core import utils as core_utils
from waldur_core.structure import executors as structure_executors
from waldur_openstack.openstack import executors as openstack_executors

from . import tasks, models


class VolumeCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return chain(
            tasks.ThrottleProvisionTask().si(
                serialized_volume,
                'create_volume',
                state_transition='begin_creating'
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error',
            ).set(countdown=30)
        )


class VolumeUpdateExecutor(core_executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        updated_fields = kwargs['updated_fields']
        if 'name' in updated_fields or 'description' in updated_fields:
            return core_tasks.BackendMethodTask().si(
                serialized_volume, 'update_volume', state_transition='begin_updating')
        else:
            return core_tasks.StateTransitionTask().si(serialized_volume, state_transition='begin_updating')


class VolumeDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        if volume.backend_id:
            return chain(
                core_tasks.BackendMethodTask().si(
                    serialized_volume, 'delete_volume', state_transition='begin_deleting'),
                core_tasks.PollBackendCheckTask().si(serialized_volume, 'is_volume_deleted'),
            )
        else:
            return core_tasks.StateTransitionTask().si(serialized_volume, state_transition='begin_deleting')


class VolumePullExecutor(core_executors.ActionExecutor):
    action = 'Pull'

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_volume, 'pull_volume',
            state_transition='begin_updating')


class VolumeExtendExecutor(core_executors.ActionExecutor):
    action = 'Extend'

    @classmethod
    def get_action_details(cls, volume, **kwargs):
        return {
            'message': 'Extend volume from %s MB to %s MB' % (kwargs.get('old_size'), volume.size),
            'old_size': kwargs.get('old_size'),
            'new_size': volume.size,
        }

    @classmethod
    def pre_apply(cls, volume, **kwargs):
        super(VolumeExtendExecutor, cls).pre_apply(volume, **kwargs)
        if volume.instance is not None:
            volume.instance.action = 'Extend volume'
            volume.instance.schedule_updating()
            volume.instance.save()

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        if volume.instance is None:
            return chain(
                core_tasks.BackendMethodTask().si(
                    serialized_volume,
                    backend_method='extend_volume',
                    state_transition='begin_updating',
                ),
                core_tasks.PollRuntimeStateTask().si(
                    serialized_volume,
                    backend_pull_method='pull_volume_runtime_state',
                    success_state='available',
                    erred_state='error'
                )
            )

        return chain(
            core_tasks.StateTransitionTask().si(
                core_utils.serialize_instance(volume.instance),
                state_transition='begin_updating'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_volume,
                backend_method='detach_volume',
                state_transition='begin_updating'
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_volume,
                backend_method='extend_volume',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_volume,
                instance_uuid=volume.instance.uuid.hex,
                device=volume.device,
                backend_method='attach_volume',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='in-use',
                erred_state='error'
            ),
        )

    @classmethod
    def get_success_signature(cls, volume, serialized_volume, **kwargs):
        if volume.instance is None:
            return super(VolumeExtendExecutor, cls).get_success_signature(volume, serialized_volume, **kwargs)
        else:
            instance = volume.instance
            serialized_instance = core_utils.serialize_instance(instance)
            return chain(
                super(VolumeExtendExecutor, cls).get_success_signature(volume, serialized_volume, **kwargs),
                super(VolumeExtendExecutor, cls).get_success_signature(instance, serialized_instance, **kwargs),
            )

    @classmethod
    def get_failure_signature(cls, volume, serialized_volume, **kwargs):
        return tasks.VolumeExtendErredTask().s(serialized_volume)


class VolumeAttachExecutor(core_executors.ActionExecutor):
    action = 'Attach'

    @classmethod
    def get_action_details(cls, volume, **kwargs):
        return {'message': 'Attach volume to instance %s' % volume.instance.name}

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_volume,
                instance_uuid=volume.instance.uuid.hex,
                device=volume.device,
                backend_method='attach_volume',
                state_transition='begin_updating'
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='in-use',
                erred_state='error',
            ),
            # additional pull to populate field "device".
            core_tasks.BackendMethodTask().si(serialized_volume, backend_method='pull_volume'),
        )


class VolumeDetachExecutor(core_executors.ActionExecutor):
    action = 'Detach'

    @classmethod
    def get_action_details(cls, volume, **kwargs):
        return {'message': 'Detach volume from instance %s' % volume.instance.name}

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_volume, backend_method='detach_volume', state_transition='begin_updating'),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error',
            )
        )


class SnapshotCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, snapshot, serialized_snapshot, **kwargs):
        return chain(
            tasks.ThrottleProvisionTask().si(
                serialized_snapshot,
                'create_snapshot',
                state_transition='begin_creating'
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_snapshot,
                backend_pull_method='pull_snapshot_runtime_state',
                success_state='available',
                erred_state='error',
            ).set(countdown=10)
        )


class SnapshotUpdateExecutor(core_executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, snapshot, serialized_snapshot, **kwargs):
        updated_fields = kwargs['updated_fields']
        # TODO: call separate task on metadata update
        if 'name' in updated_fields or 'description' in updated_fields:
            return core_tasks.BackendMethodTask().si(
                serialized_snapshot, 'update_snapshot', state_transition='begin_updating')
        else:
            return core_tasks.StateTransitionTask().si(serialized_snapshot, state_transition='begin_updating')


class SnapshotDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, snapshot, serialized_snapshot, **kwargs):
        if snapshot.backend_id:
            return chain(
                core_tasks.BackendMethodTask().si(
                    serialized_snapshot, 'delete_snapshot', state_transition='begin_deleting'),
                core_tasks.PollBackendCheckTask().si(serialized_snapshot, 'is_snapshot_deleted'),
            )
        else:
            return core_tasks.StateTransitionTask().si(serialized_snapshot, state_transition='begin_deleting')


class SnapshotPullExecutor(core_executors.ActionExecutor):
    action = 'Pull'

    @classmethod
    def get_task_signature(cls, snapshot, serialized_snapshot, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_snapshot, 'pull_snapshot',
            state_transition='begin_updating')


class InstanceCreateExecutor(core_executors.CreateExecutor):
    """ First - create instance volumes in parallel, after - create instance based on created volumes """

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, ssh_key=None, flavor=None):
        """ Create all instance volumes in parallel and wait for them to provision """
        serialized_volumes = [core_utils.serialize_instance(volume) for volume in instance.volumes.all()]

        _tasks = [tasks.ThrottleProvisionStateTask().si(serialized_instance, state_transition='begin_creating')]
        # Create volumes
        for serialized_volume in serialized_volumes:
            _tasks.append(tasks.ThrottleProvisionTask().si(
                serialized_volume, 'create_volume', state_transition='begin_creating'))
        for index, serialized_volume in enumerate(serialized_volumes):
            # Wait for volume creation
            _tasks.append(core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error',
            ).set(countdown=30 if index == 0 else 0))
            # Pull volume to sure that it is bootable
            _tasks.append(core_tasks.BackendMethodTask().si(serialized_volume, 'pull_volume'))
            # Mark volume as OK
            _tasks.append(core_tasks.StateTransitionTask().si(serialized_volume, state_transition='set_ok'))
        # Create instance based on volumes
        kwargs = {
            'backend_flavor_id': flavor.backend_id,
        }
        if ssh_key is not None:
            kwargs['public_key'] = ssh_key.public_key
        # Wait 10 seconds after volume creation due to OpenStack restrictions.
        _tasks.append(core_tasks.BackendMethodTask().si(
            serialized_instance, 'create_instance', **kwargs).set(countdown=10))

        # Wait for instance creation
        _tasks.append(core_tasks.PollRuntimeStateTask().si(
            serialized_instance,
            backend_pull_method='pull_instance_runtime_state',
            success_state=models.Instance.RuntimeStates.ACTIVE,
            erred_state=models.Instance.RuntimeStates.ERROR,
        ))

        # Update volumes runtime state and device name
        for serialized_volume in serialized_volumes:
            _tasks.append(core_tasks.BackendMethodTask().si(
                serialized_volume,
                backend_method='pull_volume',
                update_fields=['runtime_state', 'device']
            ))

        # TODO: Port should be created before instance is created.
        # The following calls should be removed: pull_created_instance_internal_ips and push_instance_internal_ips.

        # Pull instance internal IPs
        # pull_instance_internal_ips method cannot be used, because it requires backend_id to update
        # existing internal IPs. However, internal IPs of the created instance does not have backend_ids.
        _tasks.append(core_tasks.BackendMethodTask().si(serialized_instance, 'pull_created_instance_internal_ips'))

        # Consider the case when instance has several internal IPs connected to
        # different subnets within the same network.
        # When OpenStack instance is provisioned, network port is created.
        # This port is pulled into Waldur using the pull_created_instance_internal_ips method.
        # However, it does not take into account subnets, because OpenStack
        # does not allow to specify subnet on instance creation.
        # See also: https://specs.openstack.org/openstack/nova-specs/specs/juno/approved/selecting-subnet-when-creating-vm.html
        # Therefore we need to push remaining network ports for subnets explicitly.
        _tasks.append(core_tasks.BackendMethodTask().si(serialized_instance, 'push_instance_internal_ips'))

        # Pull instance security groups
        _tasks.append(core_tasks.BackendMethodTask().si(serialized_instance, 'pull_instance_security_groups'))

        # Create non-existing floating IPs
        for floating_ip in instance.floating_ips.filter(backend_id=''):
            serialized_floating_ip = core_utils.serialize_instance(floating_ip)
            _tasks.append(core_tasks.BackendMethodTask().si(serialized_floating_ip, 'create_floating_ip'))
        # Push instance floating IPs
        _tasks.append(core_tasks.BackendMethodTask().si(serialized_instance, 'push_instance_floating_ips'))
        # Wait for operation completion
        for index, floating_ip in enumerate(instance.floating_ips):
            _tasks.append(core_tasks.PollRuntimeStateTask().si(
                core_utils.serialize_instance(floating_ip),
                backend_pull_method='pull_floating_ip_runtime_state',
                success_state='ACTIVE',
                erred_state='ERRED',
            ).set(countdown=5 if not index else 0))

        shared_tenant = instance.service_project_link.service.settings.scope
        if shared_tenant:
            serialized_executor = core_utils.serialize_class(openstack_executors.TenantPullFloatingIPsExecutor)
            serialized_tenant = core_utils.serialize_instance(shared_tenant)
            _tasks.append(core_tasks.ExecutorTask().si(serialized_executor, serialized_tenant))
        return chain(*_tasks)

    @classmethod
    def get_success_signature(cls, instance, serialized_instance, **kwargs):
        return tasks.SetInstanceOKTask().si(serialized_instance)

    @classmethod
    def get_failure_signature(cls, instance, serialized_instance, **kwargs):
        return tasks.SetInstanceErredTask().s(serialized_instance)


class InstanceUpdateExecutor(core_executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        updated_fields = kwargs['updated_fields']
        if 'name' in updated_fields:
            return core_tasks.BackendMethodTask().si(
                serialized_instance, 'update_instance', state_transition='begin_updating')
        else:
            return core_tasks.StateTransitionTask().si(serialized_instance, state_transition='begin_updating')


class InstanceUpdateSecurityGroupsExecutor(core_executors.ActionExecutor):
    action = 'Update security groups'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance,
            backend_method='push_instance_security_groups',
            state_transition='begin_updating',
        )


class InstanceDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, force=False, **kwargs):
        delete_volumes = kwargs.pop('delete_volumes', True)
        release_floating_ips = kwargs.pop('release_floating_ips', True)
        delete_instance = cls.get_delete_instance_tasks(instance, serialized_instance, release_floating_ips)

        # Case 1. Instance does not exist at backend
        if not instance.backend_id:
            return core_tasks.StateTransitionTask().si(
                serialized_instance,
                state_transition='begin_deleting'
            )

        # Case 2. Instance exists at backend.
        # Data volumes are deleted by OpenStack because delete_on_termination=True
        elif delete_volumes:
            return chain(delete_instance)

        # Case 3. Instance exists at backend.
        # Data volumes are detached and not deleted.
        else:
            detach_volumes = cls.get_detach_data_volumes_tasks(instance, serialized_instance)
            return chain(detach_volumes + delete_instance)

    @classmethod
    def get_delete_instance_tasks(cls, instance, serialized_instance, release_floating_ips):
        _tasks = [
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='delete_instance',
                state_transition='begin_deleting',
            ),
            core_tasks.PollBackendCheckTask().si(
                serialized_instance,
                backend_check_method='is_instance_deleted',
            ),
        ]
        if release_floating_ips:
            for index, floating_ip in enumerate(instance.floating_ips):
                _tasks.append(core_tasks.BackendMethodTask().si(
                    core_utils.serialize_instance(floating_ip),
                    'delete_floating_ip',
                ).set(countdown=5 if not index else 0))
        else:
            # pull related floating IPs state after instance deletion
            for index, floating_ip in enumerate(instance.floating_ips):
                _tasks.append(core_tasks.BackendMethodTask().si(
                    core_utils.serialize_instance(floating_ip),
                    'pull_floating_ip_runtime_state',
                ).set(countdown=5 if not index else 0))

        shared_tenant = instance.service_project_link.service.settings.scope
        if shared_tenant:
            serialized_executor = core_utils.serialize_class(openstack_executors.TenantPullFloatingIPsExecutor)
            serialized_tenant = core_utils.serialize_instance(shared_tenant)
            _tasks.append(core_tasks.ExecutorTask().si(serialized_executor, serialized_tenant))

        return _tasks

    @classmethod
    def get_detach_data_volumes_tasks(cls, instance, serialized_instance):
        data_volumes = instance.volumes.all().filter(bootable=False)
        detach_volumes = [
            core_tasks.BackendMethodTask().si(
                core_utils.serialize_instance(volume),
                backend_method='detach_volume',
            )
            for volume in data_volumes
        ]
        check_volumes = [
            core_tasks.PollRuntimeStateTask().si(
                core_utils.serialize_instance(volume),
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error'
            )
            for volume in data_volumes
        ]
        return detach_volumes + check_volumes


class InstanceFlavorChangeExecutor(core_executors.ActionExecutor):
    action = 'Change flavor'

    @classmethod
    def get_action_details(cls, instance, **kwargs):
        old_flavor_name = kwargs.get('old_flavor_name')
        new_flavor_name = kwargs.get('flavor').name
        return {
            'message': 'Change flavor from %s to %s' % (old_flavor_name, new_flavor_name),
            'old_flavor_name': old_flavor_name,
            'new_flavor_name': new_flavor_name,
        }

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        flavor = kwargs.pop('flavor')
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='resize_instance',
                state_transition='begin_updating',
                flavor_id=flavor.backend_id
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='VERIFY_RESIZE',
                erred_state='ERRED'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='confirm_instance_resize'
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='SHUTOFF',
                erred_state='ERRED'
            ),
        )


class InstancePullExecutor(core_executors.ActionExecutor):
    action = 'Pull'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'pull_instance', state_transition='begin_updating',
            ),
            core_tasks.BackendMethodTask().si(serialized_instance, 'pull_instance_security_groups'),
            core_tasks.BackendMethodTask().si(serialized_instance, 'pull_instance_internal_ips'),
            core_tasks.BackendMethodTask().si(serialized_instance, 'pull_instance_floating_ips'),
        )


class InstanceFloatingIPsUpdateExecutor(core_executors.ActionExecutor):
    action = 'Update floating IPs'

    @classmethod
    def get_action_details(cls, instance, **kwargs):
        attached = set(instance._new_floating_ips) - set(instance._old_floating_ips)
        detached = set(instance._old_floating_ips) - set(instance._new_floating_ips)

        messages = []
        if attached:
            messages.append('Attached floating IPs: %s.' % ', '.join(attached))

        if detached:
            messages.append('Detached floating IPs: %s.' % ', '.join(detached))

        if not messages:
            messages.append('Instance floating IPs have been updated.')

        return {
            'message': ' '.join(messages),
            'attached': list(attached),
            'detached': list(detached),
        }

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        _tasks = [core_tasks.StateTransitionTask().si(serialized_instance, state_transition='begin_updating')]
        # Create non-exist floating IPs
        for floating_ip in instance.floating_ips.filter(backend_id=''):
            serialized_floating_ip = core_utils.serialize_instance(floating_ip)
            _tasks.append(core_tasks.BackendMethodTask().si(serialized_floating_ip, 'create_floating_ip'))
        # Push instance floating IPs
        _tasks.append(core_tasks.BackendMethodTask().si(serialized_instance, 'push_instance_floating_ips'))
        # Wait for operation completion
        for index, floating_ip in enumerate(instance.floating_ips):
            _tasks.append(core_tasks.PollRuntimeStateTask().si(
                core_utils.serialize_instance(floating_ip),
                backend_pull_method='pull_floating_ip_runtime_state',
                success_state='ACTIVE',
                erred_state='ERRED',
            ).set(countdown=5 if not index else 0))
        # Pull floating IPs again to update state of disconnected IPs
        _tasks.append(core_tasks.IndependentBackendMethodTask().si(serialized_instance, 'pull_floating_ips'))
        return chain(*_tasks)

    @classmethod
    def get_success_signature(cls, instance, serialized_instance, **kwargs):
        return tasks.SetInstanceOKTask().si(serialized_instance)

    @classmethod
    def get_failure_signature(cls, instance, serialized_instance, **kwargs):
        return tasks.SetInstanceErredTask().s(serialized_instance)


class InstanceStopExecutor(core_executors.ActionExecutor):
    action = 'Stop'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'stop_instance', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='SHUTOFF',
                erred_state='ERRED'
            ),
        )


class InstanceStartExecutor(core_executors.ActionExecutor):
    action = 'Start'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'start_instance', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='ACTIVE',
                erred_state='ERRED'
            ),
        )


class InstanceRestartExecutor(core_executors.ActionExecutor):
    action = 'Restart'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'restart_instance', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='ACTIVE',
                erred_state='ERRED'
            ),
        )


class InstanceInternalIPsSetUpdateExecutor(core_executors.ActionExecutor):
    action = 'Update internal IPs'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance, 'push_instance_internal_ips', state_transition='begin_updating',
        )


class BackupCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, backup, serialized_backup, **kwargs):
        serialized_snapshots = [core_utils.serialize_instance(snapshot) for snapshot in backup.snapshots.all()]

        _tasks = [core_tasks.StateTransitionTask().si(serialized_backup, state_transition='begin_creating')]
        for serialized_snapshot in serialized_snapshots:
            _tasks.append(tasks.ThrottleProvisionTask().si(
                serialized_snapshot, 'create_snapshot', force=True, state_transition='begin_creating'))
        for index, serialized_snapshot in enumerate(serialized_snapshots):
            _tasks.append(core_tasks.PollRuntimeStateTask().si(
                serialized_snapshot,
                backend_pull_method='pull_snapshot_runtime_state',
                success_state='available',
                erred_state='error',
            ).set(countdown=10 if index == 0 else 0))
            _tasks.append(core_tasks.StateTransitionTask().si(serialized_snapshot, state_transition='set_ok'))

        return chain(*_tasks)

    @classmethod
    def get_failure_signature(cls, backup, serialized_backup, **kwargs):
        return tasks.SetBackupErredTask().s(serialized_backup)


class BackupDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def pre_apply(cls, backup, **kwargs):
        for snapshot in backup.snapshots.all():
            snapshot.schedule_deleting()
            snapshot.save(update_fields=['state'])
        core_executors.DeleteExecutor.pre_apply(backup)

    @classmethod
    def get_task_signature(cls, backup, serialized_backup, force=False, **kwargs):
        serialized_snapshots = [core_utils.serialize_instance(snapshot) for snapshot in backup.snapshots.all()]

        _tasks = [core_tasks.StateTransitionTask().si(serialized_backup, state_transition='begin_deleting')]
        for serialized_snapshot in serialized_snapshots:
            _tasks.append(core_tasks.BackendMethodTask().si(
                serialized_snapshot, 'delete_snapshot', state_transition='begin_deleting'))
        for serialized_snapshot in serialized_snapshots:
            _tasks.append(core_tasks.PollBackendCheckTask().si(serialized_snapshot, 'is_snapshot_deleted'))
            _tasks.append(core_tasks.DeletionTask().si(serialized_snapshot))

        return chain(*_tasks)

    @classmethod
    def get_failure_signature(cls, backup, serialized_backup, force=False, **kwargs):
        if not force:
            return tasks.SetBackupErredTask().s(serialized_backup)
        else:
            return tasks.ForceDeleteBackupTask().si(serialized_backup)


class SnapshotRestorationExecutor(core_executors.CreateExecutor):
    """ Restores volume from snapshot instance """

    @classmethod
    def get_task_signature(cls, snapshot_restoration, serialized_snapshot_restoration, **kwargs):
        serialized_volume = core_utils.serialize_instance(snapshot_restoration.volume)

        _tasks = [
            tasks.ThrottleProvisionTask().si(
                serialized_volume, 'create_volume', state_transition='begin_creating'),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume, 'pull_volume_runtime_state', success_state='available', erred_state='error',
            ).set(countdown=30),
            core_tasks.BackendMethodTask().si(serialized_volume, 'remove_bootable_flag'),
            core_tasks.BackendMethodTask().si(serialized_volume, 'pull_volume'),
        ]

        return chain(*_tasks)

    @classmethod
    def get_success_signature(cls, snapshot_restoration, serialized_snapshot_restoration, **kwargs):
        serialized_volume = core_utils.serialize_instance(snapshot_restoration.volume)
        return core_tasks.StateTransitionTask().si(serialized_volume, state_transition='set_ok')

    @classmethod
    def get_failure_signature(cls, snapshot_restoration, serialized_snapshot_restoration, **kwargs):
        serialized_volume = core_utils.serialize_instance(snapshot_restoration.volume)
        return core_tasks.StateTransitionTask().si(serialized_volume, state_transition='set_erred')


class OpenStackTenantCleanupExecutor(structure_executors.BaseCleanupExecutor):
    related_executor = openstack_executors.OpenStackCleanupExecutor

    pre_models = (
        models.SnapshotSchedule,
        models.BackupSchedule,
    )

    executors = (
        (models.Snapshot, SnapshotDeleteExecutor),
        (models.Backup, BackupDeleteExecutor),
        (models.Instance, InstanceDeleteExecutor),
        (models.Volume, VolumeDeleteExecutor),
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations

from .. import models


def cleanup_tenant_quotas(apps, schema_editor):
    quota_names = models.Tenant.get_quotas_names()
    for obj in models.Tenant.objects.all():
        obj.quotas.exclude(name__in=quota_names).delete()


def cleanup_openstackservice_quotas(apps, schema_editor):
    quota_names = models.OpenStackService.get_quotas_names()
    for obj in models.OpenStackService.objects.all():
        obj.quotas.exclude(name__in=quota_names).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0030_subnet_dns_nameservers'),
    ]

    operations = [
        migrations.RunPython(cleanup_tenant_quotas),
        migrations.RunPython(cleanup_openstackservice_quotas),
    ]



import logging

from celery import shared_task
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import default_token_generator
from django.core.mail import send_mail
from django.template.loader import render_to_string


logger = logging.getLogger(__name__)


@shared_task(name='waldur_auth_social.send_activation_email')
def send_activation_email(user_uuid):
    user = get_user_model().objects.get(uuid=user_uuid, is_active=False)
    token = default_token_generator.make_token(user)
    url_template = settings.WALDUR_AUTH_SOCIAL['USER_ACTIVATION_URL_TEMPLATE']

    url = url_template.format(token=token, user_uuid=user_uuid)
    context = {'activation_url': url}

    subject = render_to_string('waldur_auth_social/activation_email_subject.txt', context).strip()
    text_message = render_to_string('waldur_auth_social/activation_email_message.txt', context)
    html_message = render_to_string('waldur_auth_social/activation_email_message.html', context)

    logger.debug('About to send an activation email to %s' % user.email)
    send_mail(subject, text_message, settings.DEFAULT_FROM_EMAIL, [user.email], html_message=html_message)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-20 12:17
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0014_remove_offering_preferred_language'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='item',
            new_name='orderitem'
        )
    ]



from __future__ import unicode_literals

from decimal import Decimal

from django.db import models
from django.core.validators import MinValueValidator
from django.utils.translation import ugettext_lazy as _

from waldur_core.core import models as core_models
from waldur_core.structure import models as structure_models
from waldur_mastermind.common import mixins as common_mixins


class SlurmPackage(common_mixins.ProductCodeMixin,
                   core_models.UuidMixin,
                   core_models.NameMixin,
                   models.Model):

    class Meta(object):
        verbose_name = _('SLURM package')
        verbose_name_plural = _('SLURM packages')

    class Permissions(object):
        customer_path = 'service_settings__customer'

    PRICE_MAX_DIGITS = 14
    PRICE_DECIMAL_PLACES = 10

    service_settings = models.OneToOneField(structure_models.ServiceSettings,
                                            related_name='+',
                                            limit_choices_to={'type': 'SLURM'})

    cpu_price = models.DecimalField(default=0,
                                    verbose_name=_('Price for CPU hour'),
                                    max_digits=PRICE_MAX_DIGITS,
                                    decimal_places=PRICE_DECIMAL_PLACES,
                                    validators=[MinValueValidator(Decimal('0'))])

    gpu_price = models.DecimalField(default=0,
                                    verbose_name=_('Price for GPU hour'),
                                    max_digits=PRICE_MAX_DIGITS,
                                    decimal_places=PRICE_DECIMAL_PLACES,
                                    validators=[MinValueValidator(Decimal('0'))])

    ram_price = models.DecimalField(default=0,
                                    verbose_name=_('Price for GB RAM'),
                                    max_digits=PRICE_MAX_DIGITS,
                                    decimal_places=PRICE_DECIMAL_PLACES,
                                    validators=[MinValueValidator(Decimal('0'))])



from waldur_core.core import executors as core_executors, tasks as core_tasks, utils as core_utils
from waldur_core.structure import executors as structure_executors
from waldur_openstack.openstack import executors as openstack_executors

from . import tasks


class OpenStackPackageCreateExecutor(core_executors.BaseExecutor):

    @classmethod
    def get_task_signature(cls, package, serialized_package, **kwargs):
        tenant = package.tenant
        serialized_tenant = core_utils.serialize_instance(tenant)
        service_settings = package.service_settings
        serialized_service_settings = core_utils.serialize_instance(service_settings)

        create_tenant = openstack_executors.TenantCreateExecutor.get_task_signature(tenant, serialized_tenant, **kwargs)
        set_tenant_ok = openstack_executors.TenantCreateExecutor.get_success_signature(tenant, serialized_tenant)

        populate_service_settings = tasks.OpenStackPackageSettingsPopulationTask().si(serialized_package)

        create_service_settings = structure_executors.ServiceSettingsCreateExecutor.get_task_signature(
            service_settings, serialized_service_settings)

        return create_tenant | set_tenant_ok | populate_service_settings | create_service_settings

    @classmethod
    def get_success_signature(cls, package, serialized_package, **kwargs):
        """ Get Celery signature of task that should be applied on successful execution. """
        service_settings = package.service_settings
        serialized_service_settings = core_utils.serialize_instance(service_settings)
        return core_tasks.StateTransitionTask().si(serialized_service_settings, state_transition='set_ok')

    @classmethod
    def get_failure_signature(cls, package, serialized_package, **kwargs):
        return tasks.OpenStackPackageErrorTask().s(serialized_package)


class OpenStackPackageChangeExecutor(core_executors.BaseExecutor):

    @classmethod
    def get_task_signature(cls, tenant, serialized_tenant):
        serialized_executor = core_utils.serialize_class(openstack_executors.TenantPushQuotasExecutor)
        quotas = tenant.quotas.all()
        quotas = {q.name: int(q.limit) for q in quotas}

        return core_tasks.ExecutorTask().si(serialized_executor, serialized_tenant, quotas=quotas)



from __future__ import unicode_literals

import re

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker

from model_utils.models import TimeStampedModel

from waldur_core.core import fields as core_fields
from waldur_core.core import models as core_models
from waldur_core.structure import models as structure_models
from waldur_mastermind.common import mixins as common_mixins

from . import backend, managers


@python_2_unicode_compatible
class Issue(core_models.UuidMixin,
            structure_models.StructureLoggableMixin,
            core_models.BackendModelMixin,
            TimeStampedModel,
            core_models.StateMixin):
    class Meta:
        ordering = ['-created']

    class Permissions(object):
        customer_path = 'customer'
        project_path = 'project'

    backend_id = models.CharField(max_length=255, blank=True, null=True, unique=True)
    key = models.CharField(max_length=255, blank=True)
    type = models.CharField(max_length=255)
    link = models.URLField(max_length=255, help_text=_('Link to issue in support system.'), blank=True)

    summary = models.CharField(max_length=255)
    description = models.TextField(blank=True)
    deadline = models.DateTimeField(blank=True, null=True)
    impact = models.CharField(max_length=255, blank=True)

    status = models.CharField(max_length=255)
    resolution = models.CharField(max_length=255, blank=True)
    priority = models.CharField(max_length=255, blank=True)

    caller = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='created_issues',
                               help_text=_('Waldur user who has reported the issue.'),
                               on_delete=models.PROTECT)
    reporter = models.ForeignKey('SupportUser', related_name='reported_issues', blank=True, null=True,
                                 help_text=_('Help desk user who have created the issue that is reported by caller.'),
                                 on_delete=models.PROTECT)
    assignee = models.ForeignKey('SupportUser', related_name='issues', blank=True, null=True,
                                 help_text=_('Help desk user who will implement the issue'),
                                 on_delete=models.PROTECT)

    customer = models.ForeignKey(structure_models.Customer, verbose_name=_('organization'),
                                 related_name='issues', blank=True, null=True,
                                 on_delete=models.CASCADE)
    project = models.ForeignKey(structure_models.Project, related_name='issues', blank=True, null=True,
                                on_delete=models.CASCADE)

    resource_content_type = models.ForeignKey(ContentType, null=True)
    resource_object_id = models.PositiveIntegerField(null=True)
    resource = GenericForeignKey('resource_content_type', 'resource_object_id')

    first_response_sla = models.DateTimeField(blank=True, null=True)

    tracker = FieldTracker()

    def get_description(self):
        return self.description

    @classmethod
    def get_url_name(cls):
        return 'support-issue'

    @classmethod
    def get_backend_fields(cls):
        return super(Issue, cls).get_backend_fields() + ('backend_id', 'key', 'type', 'link',
                                                         'summary', 'description', 'deadline', 'impact',
                                                         'status', 'resolution', 'priority',
                                                         'caller', 'reporter', 'assignee', 'customer', 'project',
                                                         'resource', 'first_response_sla')

    def get_log_fields(self):
        return ('uuid', 'type', 'key', 'status', 'link', 'summary',
                'reporter', 'caller', 'customer', 'project', 'resource')

    def __str__(self):
        return '{}: {}'.format(self.key or '???', self.summary)


@python_2_unicode_compatible
class SupportUser(core_models.UuidMixin, core_models.NameMixin, models.Model):
    class Meta:
        ordering = ['name']

    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='+', blank=True, null=True)
    backend_id = models.CharField(max_length=255, blank=True, null=True, unique=True)
    is_active = models.BooleanField(_('active'), default=True,
                                    help_text=_('Designates whether this user should be treated as '
                                                'active. Unselect this instead of deleting accounts.'))
    objects = managers.SupportUserManager()

    @classmethod
    def get_url_name(cls):
        return 'support-user'

    def __str__(self):
        return self.name


@python_2_unicode_compatible
class Comment(core_models.UuidMixin,
              core_models.BackendModelMixin,
              TimeStampedModel,
              core_models.StateMixin):
    class Meta:
        ordering = ['-created']
        unique_together = ('backend_id', 'issue')

    class Permissions(object):
        customer_path = 'issue__customer'
        project_path = 'issue__project'

    issue = models.ForeignKey(Issue, related_name='comments')
    author = models.ForeignKey(SupportUser, related_name='comments')
    description = models.TextField()
    is_public = models.BooleanField(default=True)
    backend_id = models.CharField(max_length=255, blank=True, null=True)

    def clean_message(self, message):
        """
        Extracts comment message from JIRA comment which contains user's info in its body.
        """
        match = re.search('^(\[.*?\]\:\s)', message)
        return message.replace(match.group(0), '') if match else message

    def prepare_message(self):
        """
        Prepends user info to the comment description to display comment author in JIRA.
        User info format - '[user.full_name user.civil_number]: '.
        """
        prefix = self.author.name
        # User is optional
        user = self.author.user
        if user:
            prefix = user.full_name or user.username
            if user.civil_number:
                prefix += ' ' + user.civil_number
        return '[%s]: %s' % (prefix, self.description)

    def update_message(self, message):
        self.description = self.clean_message(message)

    @classmethod
    def get_url_name(cls):
        return 'support-comment'

    @classmethod
    def get_backend_fields(cls):
        return super(Comment, cls).get_backend_fields() + ('issue', 'author', 'description', 'is_public', 'backend_id')

    def __str__(self):
        return self.description[:50]


@python_2_unicode_compatible
class Offering(core_models.UuidMixin,
               core_models.NameMixin,
               common_mixins.ProductCodeMixin,
               common_mixins.UnitPriceMixin,
               structure_models.StructureLoggableMixin,
               TimeStampedModel):

    class Meta:
        ordering = ['-created']
        verbose_name = _('Request')
        verbose_name_plural = _('Requests')

    class Permissions(object):
        customer_path = 'project__customer'
        project_path = 'project'

    class States(object):
        REQUESTED = 'requested'
        OK = 'ok'
        TERMINATED = 'terminated'

        CHOICES = ((REQUESTED, _('Requested')), (OK, _('OK')), (TERMINATED, _('Terminated')))

    type = models.CharField(max_length=255)
    issue = models.ForeignKey(Issue, null=True, on_delete=models.PROTECT)
    project = models.ForeignKey(structure_models.Project, null=True, on_delete=models.PROTECT)
    state = models.CharField(default=States.REQUESTED, max_length=30, choices=States.CHOICES)
    report = core_fields.JSONField(blank=True)
    terminated_at = models.DateTimeField(editable=False, blank=True, null=True)

    tracker = FieldTracker()

    def get_backend(self):
        backend.get_active_backend()

    def get_log_fields(self):
        return super(Offering, self).get_log_fields() + ('state', )

    @property
    def type_label(self):
        offerings = settings.WALDUR_SUPPORT.get('OFFERINGS', {})
        type_settings = offerings.get(self.type, {})
        return type_settings.get('label', None)

    @classmethod
    def get_url_name(cls):
        return 'support-offering'

    def __str__(self):
        return '{}: {}'.format(self.type_label or self.name, self.state)

    @classmethod
    def get_scope_type(cls):
        return 'Support.Offering'

    def _get_log_context(self, entity_name):
        context = super(Offering, self)._get_log_context(entity_name)
        context['resource_type'] = self.get_scope_type()
        context['resource_uuid'] = self.uuid.hex
        return context


@python_2_unicode_compatible
class Attachment(core_models.UuidMixin,
                 TimeStampedModel,
                 core_models.StateMixin):
    class Permissions(object):
        customer_path = 'issue__customer'
        project_path = 'issue__project'

    issue = models.ForeignKey(Issue, related_name='attachments')
    file = models.FileField(upload_to='support_attachments')
    backend_id = models.CharField(max_length=255, unique=True)
    mime_type = models.CharField(_('MIME type'), max_length=100, blank=True)
    file_size = models.PositiveIntegerField(_('Filesize, B'), blank=True, null=True)
    thumbnail = models.FileField(upload_to='support_attachments_thumbnails', blank=True, null=True)
    author = models.ForeignKey(SupportUser, related_name='attachments', blank=True, null=True)

    @classmethod
    def get_url_name(cls):
        return 'support-attachment'

    def __str__(self):
        return '{} | {}'.format(self.issue, self.file.name.split('/')[-1])


class Template(core_models.UuidMixin,
               core_models.NameMixin,
               TimeStampedModel):

    class IssueTypes(object):
        INFORMATIONAL = 'INFORMATIONAL'
        SERVICE_REQUEST = 'SERVICE_REQUEST'
        CHANGE_REQUEST = 'CHANGE_REQUEST'
        INCIDENT = 'INCIDENT'

        CHOICES = (
            (INFORMATIONAL, 'Informational'),
            (SERVICE_REQUEST, 'Service request'),
            (CHANGE_REQUEST, 'Change request'),
            (INCIDENT, 'Incident'),
        )

    native_name = models.CharField(max_length=150, blank=True)
    description = models.TextField()
    native_description = models.TextField(blank=True)
    issue_type = models.CharField(max_length=30, choices=IssueTypes.CHOICES, default=IssueTypes.INFORMATIONAL)

    @classmethod
    def get_url_name(cls):
        return 'support-template'

    def __str__(self):
        return self.name


class TemplateAttachment(core_models.UuidMixin,
                         core_models.NameMixin,
                         TimeStampedModel):
    template = models.ForeignKey(Template, on_delete=models.CASCADE, related_name='attachments')
    file = models.FileField(upload_to='support_template_attachments')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-03-23 15:33
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0014_refactor_issue_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='attachment',
            name='thumbnail',
            field=models.FileField(blank=True, null=True, upload_to='jira_attachments_thumbnails'),
        ),
    ]



from waldur_core.core import WaldurExtension


class OpenStackTenantExtension(WaldurExtension):

    class Settings:
        # wiki: https://opennode.atlassian.net/wiki/display/WD/OpenStack+plugin+configuration
        WALDUR_OPENSTACK_TENANT = {
            'MAX_CONCURRENT_PROVISION': {
                'OpenStackTenant.Instance': 4,
                'OpenStackTenant.Volume': 4,
                'OpenStackTenant.Snapshot': 4,
            },
        }

    @staticmethod
    def django_app():
        return 'waldur_openstack.openstack_tenant'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def celery_tasks():
        from datetime import timedelta
        return {
            'openstacktenant-schedule-backups': {
                'task': 'openstack_tenant.ScheduleBackups',
                'schedule': timedelta(minutes=10),
                'args': (),
            },
            'openstacktenant-delete-expired-backups': {
                'task': 'openstack_tenant.DeleteExpiredBackups',
                'schedule': timedelta(minutes=10),
                'args': (),
            },
            'openstacktenant-schedule-snapshots': {
                'task': 'openstack_tenant.ScheduleSnapshots',
                'schedule': timedelta(minutes=10),
                'args': (),
            },
            'openstacktenant-delete-expired-snapshots': {
                'task': 'openstack_tenant.DeleteExpiredSnapshots',
                'schedule': timedelta(minutes=10),
                'args': (),
            },
            'openstacktenant-set-erred-stuck-resources': {
                'task': 'openstack_tenant.SetErredStuckResources',
                'schedule': timedelta(minutes=10),
                'args': (),
            },
        }

    @staticmethod
    def get_cleanup_executor():
        from .executors import OpenStackTenantCleanupExecutor
        return OpenStackTenantCleanupExecutor



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-02-28 13:09
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import waldur_core.core.fields


class Migration(migrations.Migration):

    #replaces = [('cost_tracking', '0016_leaf_estimates'), ('cost_tracking', '0017_nullable_object_id'), ('cost_tracking', '0018_priceestimate_threshold'), ('cost_tracking', '0019_priceestimate_limit'), ('cost_tracking', '0020_reset_price_list_item'), ('cost_tracking', '0021_delete_applicationtype'), ('cost_tracking', '0022_priceestimate_leafs')]

    dependencies = [
        ('structure', '0001_squashed_0054'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('cost_tracking', '0015_defaultpricelistitem_metadata'),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='priceestimate',
            name='consumed',
            field=models.FloatField(default=0),
        ),
        migrations.AddField(
            model_name='priceestimate',
            name='scope_customer',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.Customer'),
        ),
        migrations.AlterField(
            model_name='priceestimate',
            name='content_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType'),
        ),
        migrations.AlterField(
            model_name='priceestimate',
            name='object_id',
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AddField(
            model_name='priceestimate',
            name='threshold',
            field=models.FloatField(default=0, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AddField(
            model_name='priceestimate',
            name='limit',
            field=models.FloatField(default=-1),
        ),
        migrations.AlterUniqueTogether(
            name='pricelistitem',
            unique_together=set([]),
        ),
        migrations.DeleteModel(
            name='PriceListItem',
        ),
        migrations.CreateModel(
            name='PriceListItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('value', models.DecimalField(decimal_places=5, default=0, max_digits=11, verbose_name='Hourly rate')),
                ('units', models.CharField(blank=True, max_length=255)),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('default_price_list_item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='cost_tracking.DefaultPriceListItem')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='pricelistitem',
            unique_together=set([('content_type', 'object_id', 'default_price_list_item')]),
        ),
        migrations.DeleteModel(
            name='ApplicationType',
        ),
        migrations.AddField(
            model_name='priceestimate',
            name='leafs',
            field=models.ManyToManyField(related_name='_priceestimate_leafs_+', to='cost_tracking.PriceEstimate'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-07-14 08:07
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0009_offering_article_code'),
        ('experts', '0005_expert_contract'),
    ]

    operations = [
        migrations.AddField(
            model_name='expertrequest',
            name='issue',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='support.Issue'),
        ),
    ]



from __future__ import unicode_literals

import django_filters
from django_filters.widgets import BooleanWidget

from waldur_core.core import filters as core_filters
from waldur_core.structure import filters as structure_filters

from . import models


class ServiceProjectLinkFilter(structure_filters.BaseServiceProjectLinkFilter):
    service = core_filters.URLFilter(view_name='rijkscloud-detail', name='service__uuid')

    class Meta(structure_filters.BaseServiceProjectLinkFilter.Meta):
        model = models.RijkscloudServiceProjectLink


class FlavorFilter(structure_filters.ServicePropertySettingsFilter):

    o = django_filters.OrderingFilter(fields=('cores', 'ram'))

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Flavor


class VolumeFilter(structure_filters.BaseResourceFilter):
    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Volume


class InstanceFilter(structure_filters.BaseResourceFilter):
    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Instance
        fields = structure_filters.BaseResourceFilter.Meta.fields + ('runtime_state',)


class NetworkFilter(structure_filters.ServicePropertySettingsFilter):

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Network


class SubNetFilter(structure_filters.ServicePropertySettingsFilter):
    network = core_filters.URLFilter(view_name='rijkscloud-network-detail', name='network__uuid')
    network_uuid = django_filters.UUIDFilter(name='network__uuid')

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.SubNet


class InternalIPFilter(structure_filters.ServicePropertySettingsFilter):
    is_available = django_filters.BooleanFilter(widget=BooleanWidget)

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.InternalIP
        fields = structure_filters.ServicePropertySettingsFilter.Meta.fields + (
            'address', 'subnet')


class FloatingIPFilter(structure_filters.ServicePropertySettingsFilter):
    is_available = django_filters.BooleanFilter(widget=BooleanWidget)

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.FloatingIP
        fields = structure_filters.ServicePropertySettingsFilter.Meta.fields + (
            'address',)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-06-13 07:10
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0004_offering_attributes'),
    ]

    operations = [
        migrations.AlterField(
            model_name='offering',
            name='attributes',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=''),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-05-18 15:34
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0032_nullable_internal_ip_instance'),
    ]

    operations = [
        migrations.AlterField(
            model_name='instance',
            name='backend_id',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterUniqueTogether(
            name='instance',
            unique_together=set([('service_project_link', 'backend_id')]),
        ),
        migrations.AlterField(
            model_name='snapshot',
            name='backend_id',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='volume',
            name='backend_id',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterUniqueTogether(
            name='snapshot',
            unique_together=set([('service_project_link', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='volume',
            unique_together=set([('service_project_link', 'backend_id')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-06-26 11:21
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields
import django_fsm


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0001_squashed_0054'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('marketplace', '0010_immutable_default_json'),
    ]

    operations = [
        migrations.CreateModel(
            name='Item',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('attributes', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
                ('offering', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='marketplace.Offering')),
                ('cost', models.DecimalField(blank=True, decimal_places=10, max_digits=22, null=True)),
            ],
            options={
                'verbose_name': 'Order item',
            },
        ),
        migrations.CreateModel(
            name='Order',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Project')),
                ('state', django_fsm.FSMIntegerField(choices=[(1, 'draft'), (2, 'requested for approval'), (3, 'executing'), (4, 'done'), (5, 'terminated')], default=1)),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='orders',
                                                  to=settings.AUTH_USER_MODEL)),
                ('approved_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('approved_at', models.DateTimeField(blank=True, editable=False, null=True)),
                ('total_cost', models.DecimalField(blank=True, decimal_places=10, max_digits=22, null=True)),
            ],
            options={
                'verbose_name': 'Order',
            },
        ),
        migrations.AddField(
            model_name='item',
            name='order',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='items', to='marketplace.Order'),
        ),
        migrations.AlterModelOptions(
            name='item',
            options={'ordering': ('created',), 'verbose_name': 'Order item'},
        ),
        migrations.AlterModelOptions(
            name='order',
            options={'ordering': ('created',), 'verbose_name': 'Order'},
        ),
    ]



from waldur_ansible.jupyter_hub_management import models
from waldur_core.core import models as core_models


class DeleteRequestErrorHandler(object):
    def handle_error(self, request, lines_post_processor):
        if not request.jupyter_hub_management.instance or self.exists_failed_initialization_request(request.jupyter_hub_management):
            request.jupyter_hub_management.delete()

    def exists_failed_initialization_request(self, jupyter_hub_management):
        return models.JupyterHubManagementSyncConfigurationRequest.objects.filter(jupyter_hub_management=jupyter_hub_management).latest('id').state \
            == core_models.StateMixin.States.ERRED



from waldur_core.cost_tracking import CostTrackingStrategy, ConsumableItem, CostTrackingRegister

from . import models, backend


class AzureCostTrackingStrategy(CostTrackingStrategy):
    resource_class = models.VirtualMachine

    class Types(object):
        FLAVOR = 'flavor'

    @classmethod
    def get_consumable_items(cls):
        return [ConsumableItem(item_type=cls.Types.FLAVOR, key=size.name, default_price=size.price)
                for size in backend.SizeQueryset()]

    @classmethod
    def get_configuration(cls, virtual_machine):
        consumables = {}
        if virtual_machine.state != models.VirtualMachine.States.ERRED and virtual_machine.image_name:
            consumables[ConsumableItem(item_type=cls.Types.FLAVOR, key=virtual_machine.image_name)] = 1
        return consumables


CostTrackingRegister.register_strategy(AzureCostTrackingStrategy)



from __future__ import unicode_literals

from waldur_core.core import WaldurExtension


class MarketplaceExtension(WaldurExtension):
    class Settings:
        WALDUR_MARKETPLACE = {
            'THUMBNAIL_SIZE': (120, 120),
            'THUMBNAIL_SUFFIX': 'thumbnail',
            'OWNER_CAN_APPROVE_ORDER': True,
            'MANAGER_CAN_APPROVE_ORDER': False,
            'ADMIN_CAN_APPROVE_ORDER': False,
        }

    @staticmethod
    def django_app():
        return 'waldur_mastermind.marketplace'

    @staticmethod
    def is_assembly():
        return True

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def celery_tasks():
        return {}



from __future__ import unicode_literals

from celery import chain

from waldur_core.core import executors as core_executors, tasks as core_tasks
from waldur_core.structure import executors as structure_executors

from . import models


class VirtualMachineStartExecutor(core_executors.ActionExecutor):
    action = 'Start'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, backend_method='start_vm', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_virtual_machine_runtime_state',
                success_state='running',
                erred_state='erred'
            ),
        )


class VirtualMachineStopExecutor(core_executors.ActionExecutor):
    action = 'Stop'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, backend_method='stop_vm', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_virtual_machine_runtime_state',
                success_state='stopped',
                erred_state='error'
            ),
        )


class VirtualMachineRestartExecutor(core_executors.ActionExecutor):
    action = 'Restart'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, backend_method='reboot_vm', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_virtual_machine_runtime_state',
                success_state='running',
                erred_state='error'
            ),
        )


class VirtualMachineCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='provision_vm',
                state_transition='begin_creating',
                **kwargs
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_virtual_machine_runtime_state',
                success_state='running',
                erred_state='error',
            ).set(countdown=30),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='pull_vm_info',
            ),
        )


class VirtualMachineDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id:
            return chain(
                core_tasks.BackendMethodTask().si(
                    serialized_instance, backend_method='destroy_vm', state_transition='begin_deleting'),
                core_tasks.PollBackendCheckTask().si(serialized_instance, 'is_vm_deleted'),
            )
        else:
            return core_tasks.StateTransitionTask().si(serialized_instance, state_transition='begin_deleting')


class AzureCleanupExecutor(structure_executors.BaseCleanupExecutor):
    executors = (
        (models.VirtualMachine, VirtualMachineDeleteExecutor),
    )



from __future__ import unicode_literals

from django.conf.urls import url

from waldur_mastermind.billing import views


def register_in(router):
    router.register(r'billing-price-estimates', views.PriceEstimateViewSet, base_name='billing-price-estimate')


urlpatterns = [
    url(r'^api/billing-total-cost/$', views.TotalCustomerCostView.as_view()),
]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0040_make_is_active_nullable'),
    ]

    operations = [
        migrations.AddField(
            model_name='servicesettings',
            name='domain',
            field=models.CharField(max_length=200, null=True, blank=True),
        ),
    ]




def copy_cloud_service_name_on_service_creation(sender, instance, created=False, **kwargs):
    if not created:
        return

    service_project_link = instance
    if instance.cloud_service_name or 'cloud_service_name' not in service_project_link.service.settings.options:
        return

    cloud_service_name = service_project_link.service.settings.options['cloud_service_name']
    instance.cloud_service_name = cloud_service_name
    instance.save()



from django.core.management.base import BaseCommand

from waldur_core.logging import loggers
from waldur_core.logging.models import EventTypesMixin


class Command(BaseCommand):
    help = "Cleanup stale event types in all hooks."

    def handle(self, *args, **options):
        self.stdout.write('Checking event types of hooks...')

        valid_events = loggers.get_valid_events()
        changed_hooks = 0
        for model in EventTypesMixin.get_all_models():
            for hook in model.objects.all():
                clean_events = filter(lambda x: x in valid_events, hook.event_types)
                if clean_events != hook.event_types:
                    hook.event_types = clean_events
                    hook.save(update_fields=['event_types'])
                    changed_hooks += 1

        if changed_hooks == 0:
            self.stdout.write('All hooks have valid event types.')
        elif changed_hooks == 1:
            self.stdout.write('1 hook has been updated.')
        else:
            self.stdout.write('%s hooks have been updated.' % changed_hooks)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0019_remove_payable_mixin'),
    ]

    operations = [
        migrations.AddField(
            model_name='tenant',
            name='extra_configuration',
            field=waldur_core.core.fields.JSONField(default={}, help_text='Configuration details that are not represented on backend.'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.db.models import Count


def remove_duplicates(apps, schema_editor):
    SecurityGroup = apps.get_model('openstack', 'SecurityGroup')
    OpenStackServiceProjectLink = apps.get_model('openstack', 'OpenStackServiceProjectLink')
    OK_STATE = 3

    for spl in OpenStackServiceProjectLink.objects.iterator():
        spl_security_groups = SecurityGroup.objects.filter(service_project_link=spl)
        for duplicated_backend_id in spl_security_groups.values_list('backend_id', flat=True).annotate(
                duplicates=Count('backend_id')).filter(duplicates__gt=1):

            duplicates_query = spl_security_groups.filter(backend_id=duplicated_backend_id)

            # try to leave only support groups in OK state
            if duplicates_query.filter(state=OK_STATE).count() > 0:
                security_group = duplicates_query.filter(state=OK_STATE).first()
            else:
                security_group = duplicates_query.first()

            duplicates_query.exclude(id=security_group.id).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0035_remove_ipmapping'),
    ]

    operations = [
        migrations.RunPython(remove_duplicates),
    ]



from __future__ import unicode_literals

from datetime import timedelta

from waldur_core.core import WaldurExtension


class SupportExtension(WaldurExtension):
    class Settings(object):
        WALDUR_SUPPORT = {
            # wiki for global options: https://opennode.atlassian.net/wiki/display/WD/Assembly+plugin+configuration
            'ENABLED': False,
            # wiki for JIRA-specific options:
            # https://opennode.atlassian.net/wiki/display/WD/JIRA+Service+Desk+configuration
            'ACTIVE_BACKEND': 'waldur_mastermind.support.backend.atlassian:ServiceDeskBackend',
            'CREDENTIALS': {
                'server': 'http://example.com/',
                'username': 'USERNAME',
                'password': 'PASSWORD',
                'verify_ssl': False,
            },
            'PROJECT': {
                'key': 'PROJECT',
            },
            'ISSUE': {
                'types': ['Informational', 'Service Request', 'Change Request', 'Incident'],
                'impact_field': 'Impact',
                'reporter_field': 'Original Reporter',
                'caller_field': 'Caller',
                'sla_field': 'Time to first response',
                'summary': '{{issue.summary}}',
                'description': (
                    '{{issue.description}}\n\n'
                    '---\n'
                    'Additional Info: \n'
                    '{% if issue.customer %}'
                    '- Organization: {{issue.customer.name}}\n'
                    '{% endif %}'
                    '{% if issue.project %}'
                    '- Project: {{issue.project.name}}\n'
                    '{% endif %}'
                    '{% if issue.resource %}'
                    '{% if issue.resource.service_project_link and issue.resource.service_project_link.service %}'
                    '{% if issue.resource.service_project_link.service.type %}'
                    '- Service type: {{issue.resource.service_project_link.service.type}}\n'
                    '{% endif %}'
                    '{% endif %}'
                    '- Affected resource: {{issue.resource}}\n'
                    '{% endif %}'
                ),
            },
            'DEFAULT_OFFERING_ISSUE_TYPE': 'Service Request',
            'TERMINATED_OFFERING_LIFETIME': timedelta(weeks=2),
            'OFFERINGS': {
                # An example of configuration for debugging purposes.
                # Add it to settings file to enable Custom VPC offering
                # 'custom_vpc': {
                #     'label': 'Custom VPC',
                #     'order': ['storage', 'ram', 'cpu_count'],
                #     'icon': 'fa-gear',
                #     'category': 'Custom requests',
                #     'description': 'Custom VPC example.',
                #     'options': {
                #         'storage': {
                #             'type': 'integer',
                #             'label': 'Max storage, GB',
                #             'required': True,
                #             'help_text': 'VPC storage limit in GB.',
                #         },
                #         'ram': {
                #             'type': 'integer',
                #             'label': 'Max RAM, GB',
                #             'required': True,
                #             'help_text': 'VPC RAM limit in GB.',
                #         },
                #         'cpu_count': {
                #             'type': 'integer',
                #             'label': 'Max vCPU',
                #             'required': True,
                #             'help_text': 'VPC CPU count limit.',
                #         },
                #     },
                # },
            },
            'EXCLUDED_ATTACHMENT_TYPES': [],
        }

        SUPPRESS_NOTIFICATION_EMAILS = False
        ISSUE_LINK_TEMPLATE = 'https://www.example.com/#/support/issue/{uuid}/'

    @staticmethod
    def django_app():
        return 'waldur_mastermind.support'

    @staticmethod
    def django_urls():
        from .urls import urlpatterns
        return urlpatterns

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def is_assembly():
        return True

    @staticmethod
    def celery_tasks():
        return {
            'pull-support-users': {
                'task': 'support.SupportUserPullTask',
                'schedule': timedelta(hours=6),
                'args': (),
            },
            'remove-terminated-offerings': {
                'task': 'waldur_mastermind.support.remove_terminated_offerings',
                'schedule': timedelta(hours=24),
                'args': (),
            },
        }

    @staticmethod
    def get_cleanup_executor():
        from waldur_mastermind.support.executors import SupportCleanupExecutor
        return SupportCleanupExecutor



"""
Business logic of invoice item registration.

Glossary:
 - item - invoice item. Part of the invoice that stores price for
          some object that was bought by customer.
 - source - invoice item source. Object that was bought by customer.

Example: Offering (source) ->  OfferingItem (item).

RegistrationManager represents the highest level of business logic and should be
used for invoice items registration and termination.
Registrators defines items creation and termination logic for each invoice item.
"""
from django.db import transaction
from django.utils import timezone

from waldur_core.core import utils as core_utils


class BaseRegistrator(object):

    def get_customer(self, source):
        """ Return customer based on provided item. """
        raise NotImplementedError()

    def register(self, sources, invoice, start):
        """ For each source create invoice item and register it in invoice. """
        end = core_utils.month_end(start)
        for source in sources:
            self._create_item(source, invoice, start=start, end=end)

    def get_sources(self, customer):
        """ Return a list of invoice item sources to charge customer for. """
        raise NotImplementedError()

    def _create_item(self, source, invoice, start, end):
        """ Register single chargeable item in the invoice. """
        raise NotImplementedError()

    def terminate(self, source, now=None):
        """
        Freeze invoice item's usage.
        :param source: chargeable item to use for search of invoice item.
        :param now: date of invoice with invoice items.
        """
        if not now:
            now = timezone.now()

        item = self._find_item(source, now)
        if item:
            item.terminate(end=now)

    def _find_item(self, source, now):
        """
        Find item by source and date.
        :param source: object that was bought by customer.
        :param now: date of invoice with invoice items.
        :return: invoice item or None
        """
        raise NotImplementedError()

    def get_name(self, source):
        return source.name

    def get_details(self, source):
        return {}


class RegistrationManager(object):
    """ The highest interface for invoice item registration and termination. """
    _registrators = {}

    @classmethod
    def get_registrators(cls):
        return cls._registrators.values()

    @classmethod
    def add_registrator(cls, model, registrator):
        cls._registrators[model] = registrator()

    @classmethod
    def get_registrator(cls, source):
        return cls._registrators[source.__class__]

    @classmethod
    def get_or_create_invoice(cls, customer, date):
        from . import models
        invoice, created = models.Invoice.objects.get_or_create(
            customer=customer,
            month=date.month,
            year=date.year,
        )

        if created:
            for registrator in cls.get_registrators():
                sources = registrator.get_sources(customer)
                registrator.register(sources, invoice, date)

        return invoice, created

    @classmethod
    def register(cls, source, now=None):
        """
        Create new invoice item from source and register it into invoice.

        If invoice does not exist new one will be created.
        """
        if now is None:
            now = timezone.now()

        registrator = cls._registrators[source.__class__]
        customer = registrator.get_customer(source)

        with transaction.atomic():
            invoice, created = cls.get_or_create_invoice(customer, now)
            if not created:
                registrator.register([source], invoice, now)

    @classmethod
    def terminate(cls, source, now=None):
        """
        Terminate invoice item that corresponds given source.

        :param now: time to set as end of item usage.
        """
        if now is None:
            now = timezone.now()

        registrator = cls._registrators[source.__class__]
        customer = registrator.get_customer(source)

        with transaction.atomic():
            cls.get_or_create_invoice(customer, now)
            registrator.terminate(source, now)

    @classmethod
    def get_item(cls, source, now=None):
        if now is None:
            now = timezone.now()

        registrator = cls._registrators[source.__class__]
        return registrator._find_item(source, now)

    @classmethod
    def get_name(cls, source):
        registrator = cls._registrators[source.__class__]
        return registrator.get_name(source)

    @classmethod
    def get_details(cls, source):
        registrator = cls._registrators[source.__class__]
        return registrator.get_details(source)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import waldur_core.structure.models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0026_add_error_message'),
    ]

    operations = [
        migrations.AddField(
            model_name='servicesettings',
            name='service_type',
            field=models.CharField(max_length=255, default='', db_index=True, validators=[waldur_core.structure.models.validate_service_type]),
            preserve_default=True,
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-29 12:41
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0052_customer_subnets'),
        ('waldur_jira', '0007_add_priority'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='issuetype',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='priority',
            unique_together=set([('settings', 'backend_id')]),
        ),
    ]



class PythonManagementConstants(object):
    INSTALL_PYTHON_ENVIRONMENT = 'install_python_environment'
    SYNCHRONIZE_PACKAGES = 'synchronize_packages'
    FIND_INSTALLED_LIBRARIES_FOR_VIRTUAL_ENVIRONMENT = 'find_installed_libraries_for_virtual_environment'
    FIND_INSTALLED_VIRTUAL_ENVIRONMENTS = 'find_installed_virtual_environments'
    DELETE_VIRTUAL_ENVIRONMENT = 'delete_virtual_environment'
    DELETE_PYTHON_ENVIRONMENT = 'delete_python_environment'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-09-14 09:12
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0029_add_unique_constraint_for_properties'),
    ]

    operations = [
        migrations.AddField(
            model_name='volume',
            name='image_name',
            field=models.CharField(blank=True, max_length=150),
        ),
        migrations.AlterField(
            model_name='volume',
            name='image',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='openstack_tenant.Image'),
        ),
    ]



from __future__ import absolute_import

import abc
import logging
import subprocess  # nosec

from django.utils.functional import cached_property
import six

from .structures import Quotas


logger = logging.getLogger(__name__)


class BatchError(Exception):
    pass


@six.add_metaclass(abc.ABCMeta)
class BaseBatchClient(object):

    def __init__(self, hostname, key_path, username='root', port=22, use_sudo=False):
        self.hostname = hostname
        self.key_path = key_path
        self.username = username
        self.port = port
        self.use_sudo = use_sudo

    @abc.abstractmethod
    def list_accounts(self):
        """
        Get accounts list.
        :return: list[structures.Account object]
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_account(self, name):
        """
        Get account info.
        :param name: [string] batch account name
        :return: [structures.Account object]
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def create_account(self, name, description, organization, parent_name=None):
        """
        Create account.
        :param name: [string] account name
        :param description: [string] account description
        :param organization: [string] account organization name
        :param parent_name: [string] account parent name. Optional.
        :return: None
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def delete_account(self, name):
        """
        Delete account.
        :param name: [string] account name
        :return: None
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def set_resource_limits(self, account, quotas):
        """
        Set account limits.
        :param account: [string] account name
        :param quotas: [structures.Quotas object] limits
        :return: None
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_association(self, user, account):
        """
        Get association user and account.
        :param user: [string] user name
        :param account: [string] account name
        :return: [structures.Association object]
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def create_association(self, username, account, default_account=None):
        """
        Create association user and account
        :param username: [string] user name
        :param account: [string] account name
        :param default_account: [string] default account name. Optional.
        :return: None
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def delete_association(self, username, account):
        """
        Delete_association user and account.
        :param username: [string] user name
        :param account: [string] account name
        :return: None
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_usage_report(self, accounts):
        """
        Get usages records.
        :param accounts: list[string]
        :return: list[BaseReportLine]
        """
        raise NotImplementedError()

    def execute_command(self, command):
        server = '%s@%s' % (self.username, self.hostname)
        port = str(self.port)
        if self.use_sudo:
            account_command = ['sudo']
        else:
            account_command = []

        account_command.extend(command)
        ssh_command = ['ssh', '-o', 'UserKnownHostsFile=/dev/null', '-o', 'StrictHostKeyChecking=no',
                       server, '-p', port, '-i', self.key_path, ' '.join(account_command)]
        try:
            logger.debug('Executing SSH command: %s', ' '.join(ssh_command))
            return subprocess.check_output(ssh_command, stderr=subprocess.STDOUT)  # nosec
        except subprocess.CalledProcessError as e:
            logger.exception('Failed to execute command "%s".', ssh_command)
            stdout = e.output or ''
            lines = stdout.splitlines()
            if len(lines) > 0 and lines[0].startswith('Warning: Permanently added'):
                lines = lines[1:]
            stdout = '\n'.join(lines)
            six.reraise(BatchError, stdout)


@six.add_metaclass(abc.ABCMeta)
class BaseReportLine(object):
    @abc.abstractproperty
    def account(self):
        pass

    @abc.abstractproperty
    def user(self):
        pass

    @property
    def cpu(self):
        return 0

    @property
    def gpu(self):
        return 0

    @property
    def ram(self):
        return 0

    @property
    def duration(self):
        return 0

    @property
    def charge(self):
        return 0

    @property
    def node(self):
        return 0

    @cached_property
    def quotas(self):
        return Quotas(
            self.cpu * self.duration * self.node,
            self.gpu * self.duration * self.node,
            self.ram * self.duration * self.node,
            self.charge
        )



from django.conf.urls import url
from djangosaml2.views import metadata

from .views import Saml2LoginView, Saml2LoginCompleteView, Saml2LogoutView, Saml2LogoutCompleteView, Saml2ProviderView


urlpatterns = [
    url(r'^api-auth/saml2/login/complete/$', Saml2LoginCompleteView.as_view(), name='saml2-login-complete'),
    url(r'^api-auth/saml2/login/$', Saml2LoginView.as_view(), name='saml2-login'),
    url(r'^api-auth/saml2/logout/complete/$', Saml2LogoutCompleteView.as_view(), name='saml2-logout-complete'),
    url(r'^api-auth/saml2/logout/$', Saml2LogoutView.as_view(), name='saml2-logout'),
    url(r'^api-auth/saml2/metadata/$', metadata, name='saml2-metadata'),
    url(r'^api-auth/saml2/providers/$', Saml2ProviderView.as_view(), name='saml2-provider'),
]



from __future__ import unicode_literals

from waldur_core.core import tasks as core_tasks
from waldur_openstack.openstack import models as openstack_models


class OpenStackPackageErrorTask(core_tasks.ErrorStateTransitionTask):
    """ Handle OpenStack package creation errors.

        If error occurred on tenant creation - mark tenant and service settings as erred.
        If error occurred on service settings creation - mark only service settings as erred.
    """

    @classmethod
    def get_description(cls, result_id, package, *args, **kwargs):
        return 'Mark package "%s" components as erred.' % package

    def execute(self, package):
        if package.tenant.state != openstack_models.Tenant.States.OK:
            self.state_transition(package.tenant, 'set_erred')
            self.save_error_message(package.tenant)
            self.state_transition(package.service_settings, 'set_erred')
            package.service_settings.error_message = 'Failed to create tenant: %s.' % package.tenant
            package.service_settings.save(update_fields=['error_message'])
        else:
            self.state_transition(package.service_settings, 'set_erred')
            self.save_error_message(package.service_settings)


class OpenStackPackageSettingsPopulationTask(core_tasks.Task):
    """ Populate service settings options based on provisioned tenant. """

    @classmethod
    def get_description(cls, package, *args, **kwargs):
        return 'Copy tenant backend data to settings. Package "%s".' % package

    def execute(self, package):
        package.service_settings.options['tenant_id'] = package.tenant.backend_id
        package.service_settings.options['external_network_id'] = package.tenant.external_network_id
        package.service_settings.options['internal_network_id'] = package.tenant.internal_network_id
        package.service_settings.save()



from __future__ import unicode_literals

import collections
import json
import logging
from datetime import datetime

import dateutil.parser
from django.conf import settings
from django.utils import timezone
from jira import Comment
from jira.utils import json_loads
from six.moves.html_parser import HTMLParser

from waldur_jira.backend import reraise_exceptions, JiraBackend
from waldur_mastermind.support import models

from . import SupportBackend

logger = logging.getLogger(__name__)


Settings = collections.namedtuple('Settings', ['backend_url', 'username', 'password'])


class ServiceDeskBackend(JiraBackend, SupportBackend):
    servicedeskapi_path = 'servicedeskapi'
    model_comment = models.Comment
    model_issue = models.Issue
    model_attachment = models.Attachment

    def __init__(self):
        self.settings = Settings(
            backend_url=settings.WALDUR_SUPPORT.get('CREDENTIALS', {}).get('server'),
            username=settings.WALDUR_SUPPORT.get('CREDENTIALS', {}).get('username'),
            password=settings.WALDUR_SUPPORT.get('CREDENTIALS', {}).get('password'),
        )
        self.verify = settings.WALDUR_SUPPORT.get('CREDENTIALS', {}).get('verify_ssl')
        self.project_settings = settings.WALDUR_SUPPORT.get('PROJECT', {})
        self.issue_settings = settings.WALDUR_SUPPORT.get('ISSUE', {})

    @reraise_exceptions
    def create_comment(self, comment):
        backend_comment = self._add_comment(
            comment.issue.backend_id,
            comment.prepare_message(),
            is_internal=not comment.is_public,
        )
        comment.backend_id = backend_comment.id
        comment.save(update_fields=['backend_id'])

    def _add_comment(self, issue, body, is_internal):
        data = {
            'body': body,
            'properties': [{'key': 'sd.public.comment', 'value': {'internal': is_internal}}, ]
        }

        url = self.manager._get_url('issue/{0}/comment'.format(issue))
        response = self.manager._session.post(url, data=json.dumps(data))

        comment = Comment(self.manager._options, self.manager._session, raw=json_loads(response))
        return comment

    @reraise_exceptions
    def create_issue(self, issue):
        if not issue.caller.email:
            return

        self.create_user(issue.caller)
        super(ServiceDeskBackend, self).create_issue(issue)

    def create_user(self, user):
        return self.manager.add_user(user.email, user.email, fullname=user.full_name, ignore_existing=True)

    @reraise_exceptions
    def get_users(self):
        users = self.manager.search_assignable_users_for_projects('', self.project_settings['key'], maxResults=False)
        return [models.SupportUser(name=user.displayName, backend_id=user.key) for user in users]

    def _issue_to_dict(self, issue):
        parser = HTMLParser()
        args = {
            'project': self.project_settings['key'],
            'summary': parser.unescape(issue.summary),
            'description': parser.unescape(issue.description),
            'issuetype': {'name': issue.type},
        }

        if issue.reporter:
            args[self.get_field_id_by_name(self.issue_settings['reporter_field'])] = issue.reporter.name
        if issue.impact:
            args[self.get_field_id_by_name(self.issue_settings['impact_field'])] = issue.impact
        if issue.priority:
            args['priority'] = {'name': issue.priority}

        args[self.get_field_id_by_name(self.issue_settings['caller_field'])] = [{
            "name": issue.caller.email,
            "key": issue.caller.email
        }]
        return args

    def _get_first_sla_field(self, backend_issue):
        field_name = self.get_field_id_by_name(self.issue_settings['sla_field'])
        value = getattr(backend_issue.fields, field_name, None)
        if value and hasattr(value, 'ongoingCycle'):
            epoch_milliseconds = value.ongoingCycle.breachTime.epochMillis
            if epoch_milliseconds:
                return datetime.fromtimestamp(epoch_milliseconds / 1000.0, timezone.get_default_timezone())

    def _backend_issue_to_issue(self, backend_issue, issue):
        issue.key = backend_issue.key
        issue.backend_id = backend_issue.key
        issue.resolution = backend_issue.fields.resolution or ''
        issue.status = backend_issue.fields.status.name or ''
        issue.link = backend_issue.permalink()
        issue.priority = backend_issue.fields.priority.name
        issue.first_response_sla = self._get_first_sla_field(backend_issue)
        issue.summary = backend_issue.fields.summary
        issue.description = backend_issue.fields.description or ''
        issue.type = backend_issue.fields.issuetype.name

        def get_support_user_by_field(fields, field_name):
            support_user = None
            backend_user = getattr(fields, field_name, None)

            if backend_user:
                support_user_backend_key = getattr(backend_user, 'key', None)

                if support_user_backend_key:
                    support_user, _ = models.SupportUser.objects.get_or_create(backend_id=support_user_backend_key)

            return support_user

        impact_field_id = self.get_field_id_by_name(self.issue_settings['impact_field'])
        impact = getattr(backend_issue.fields, impact_field_id, None)
        if impact:
            issue.impact = impact

        assignee = get_support_user_by_field(backend_issue.fields, 'assignee')
        if assignee:
            issue.assignee = assignee

        reporter = get_support_user_by_field(backend_issue.fields, 'reporter')
        if reporter:
            issue.reporter = reporter

    def _backend_comment_to_comment(self, backend_comment, comment):
        comment.update_message(backend_comment.body)
        author, _ = models.SupportUser.objects.get_or_create(backend_id=backend_comment.author.key)
        comment.author = author
        internal = self._get_property('comment', backend_comment.id, 'sd.public.comment')
        comment.is_public = not internal.get('value', {}).get('internal', False)

    def _backend_attachment_to_attachment(self, backend_attachment, attachment):
        author, _ = models.SupportUser.objects.get_or_create(backend_id=backend_attachment.author.key)
        attachment.mime_type = getattr(backend_attachment, 'mimeType', '')
        attachment.file_size = backend_attachment.size
        attachment.created = dateutil.parser.parse(backend_attachment.created)
        attachment.author = author



from __future__ import unicode_literals

import django_filters
from django.contrib import auth

from . import models

User = auth.get_user_model()


class ApplicationFilter(django_filters.FilterSet):
    project = django_filters.UUIDFilter(name='project__uuid')
    project_uuid = django_filters.UUIDFilter(name='project__uuid')
    project_name = django_filters.CharFilter(name='project__name', lookup_expr='icontains')

    class Meta(object):
        model = models.ApplicationModel
        fields = []



from __future__ import unicode_literals

from django.db import transaction

from . import tasks


def create_screenshot_thumbnail(sender, instance, created=False, **kwargs):
    if not created:
        return

    transaction.on_commit(lambda: tasks.create_screenshot_thumbnail.delay(instance.uuid))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from decimal import Decimal

import django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0006_offering'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='offering',
            name='type_label',
        ),
        migrations.RemoveField(
            model_name='offering',
            name='description',
        ),
        migrations.AlterField(
            model_name='offering',
            name='type',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='offering',
            name='price',
            field=models.DecimalField(decimal_places=7, default=0, max_digits=13, validators=[django.core.validators.MinValueValidator(Decimal('0'))], help_text='Price per day', verbose_name='Price per day'),
        ),
    ]



from __future__ import unicode_literals

from django.db import models
from django.template.defaultfilters import slugify
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from waldur_core.core.fields import JSONField
from model_utils import FieldTracker
from model_utils.models import TimeStampedModel

from six.moves import urllib
from waldur_core.core import models as core_models, utils as core_utils
from waldur_core.logging.loggers import LoggableMixin
from waldur_core.quotas.fields import QuotaField, UsageAggregatorQuotaField, CounterQuotaField
from waldur_core.quotas.models import QuotaModelMixin
from waldur_core.structure import models as structure_models

from waldur_openstack.openstack_base import models as openstack_base_models


class ServiceUsageAggregatorQuotaField(UsageAggregatorQuotaField):
    def __init__(self, **kwargs):
        super(ServiceUsageAggregatorQuotaField, self).__init__(
            get_children=lambda service: Tenant.objects.filter(
                service_project_link__service=service
            ), **kwargs)


class OpenStackService(structure_models.Service):
    projects = models.ManyToManyField(
        structure_models.Project, related_name='openstack_services', through='OpenStackServiceProjectLink')

    class Meta:
        unique_together = ('customer', 'settings')
        verbose_name = _('OpenStack provider')
        verbose_name_plural = _('OpenStack providers')

    class Quotas(QuotaModelMixin.Quotas):
        tenant_count = CounterQuotaField(
            target_models=lambda: [Tenant],
            path_to_scope='service_project_link.service'
        )
        vcpu = ServiceUsageAggregatorQuotaField()
        ram = ServiceUsageAggregatorQuotaField()
        storage = ServiceUsageAggregatorQuotaField()
        instances = ServiceUsageAggregatorQuotaField()
        security_group_count = ServiceUsageAggregatorQuotaField()
        security_group_rule_count = ServiceUsageAggregatorQuotaField()
        floating_ip_count = ServiceUsageAggregatorQuotaField()
        volumes = ServiceUsageAggregatorQuotaField()
        snapshots = ServiceUsageAggregatorQuotaField()

    @classmethod
    def get_url_name(cls):
        return 'openstack'


class OpenStackServiceProjectLink(structure_models.ServiceProjectLink):

    service = models.ForeignKey(OpenStackService)

    class Meta(structure_models.ServiceProjectLink.Meta):
        verbose_name = _('OpenStack provider project link')
        verbose_name_plural = _('OpenStack provider project links')

    @classmethod
    def get_url_name(cls):
        return 'openstack-spl'

    # XXX: Hack for statistics: return quotas of tenants as quotas of SPLs.
    @classmethod
    def get_sum_of_quotas_as_dict(cls, spls, quota_names=None, fields=['usage', 'limit']):
        tenants = Tenant.objects.filter(service_project_link__in=spls)
        return Tenant.get_sum_of_quotas_as_dict(tenants, quota_names=quota_names, fields=fields)


class Flavor(LoggableMixin, structure_models.ServiceProperty):
    cores = models.PositiveSmallIntegerField(help_text=_('Number of cores in a VM'))
    ram = models.PositiveIntegerField(help_text=_('Memory size in MiB'))
    disk = models.PositiveIntegerField(help_text=_('Root disk size in MiB'))

    @classmethod
    def get_url_name(cls):
        return 'openstack-flavor'

    @classmethod
    def get_backend_fields(cls):
        readonly_fields = super(Flavor, cls).get_backend_fields()
        return readonly_fields + ('cores', 'ram', 'disk')


class Image(openstack_base_models.BaseImage):
    @classmethod
    def get_url_name(cls):
        return 'openstack-image'


class SecurityGroup(structure_models.SubResource):
    service_project_link = models.ForeignKey(
        OpenStackServiceProjectLink, related_name='security_groups')
    tenant = models.ForeignKey('Tenant', related_name='security_groups')

    def get_backend(self):
        return self.tenant.get_backend()

    @classmethod
    def get_url_name(cls):
        return 'openstack-sgp'

    def increase_backend_quotas_usage(self, validate=True):
        self.tenant.add_quota_usage(self.tenant.Quotas.security_group_count, 1, validate=validate)
        self.tenant.add_quota_usage(self.tenant.Quotas.security_group_rule_count, self.rules.count(), validate=validate)

    def decrease_backend_quotas_usage(self):
        self.tenant.add_quota_usage(self.tenant.Quotas.security_group_count, -1)
        self.tenant.add_quota_usage(self.tenant.Quotas.security_group_rule_count, -self.rules.count())

    def change_backend_quotas_usage_on_rules_update(self, old_rules_count, validate=True):
        count = self.rules.count() - old_rules_count
        self.tenant.add_quota_usage(self.tenant.Quotas.security_group_rule_count, count, validate=validate)

    @classmethod
    def get_backend_fields(cls):
        return super(SecurityGroup, cls).get_backend_fields() + ('name', 'description')


class SecurityGroupRule(openstack_base_models.BaseSecurityGroupRule):
    security_group = models.ForeignKey(SecurityGroup, related_name='rules')


@python_2_unicode_compatible
class FloatingIP(core_models.RuntimeStateMixin, structure_models.SubResource):
    service_project_link = models.ForeignKey(
        OpenStackServiceProjectLink, related_name='floating_ips')
    tenant = models.ForeignKey('Tenant', related_name='floating_ips')
    address = models.GenericIPAddressField(null=True, blank=True, protocol='IPv4', default=None)
    backend_network_id = models.CharField(max_length=255, editable=False)

    tracker = FieldTracker()

    class Meta:
        unique_together = ('tenant', 'address')
        verbose_name = _('Floating IP')
        verbose_name_plural = _('Floating IPs')

    def get_backend(self):
        return self.tenant.get_backend()

    @classmethod
    def get_url_name(cls):
        return 'openstack-fip'

    def __str__(self):
        return '%s:%s (%s)' % (self.address, self.runtime_state, self.service_project_link)

    @classmethod
    def get_backend_fields(cls):
        return super(FloatingIP, cls).get_backend_fields() + ('name', 'description', 'address', 'backend_network_id',
                                                              'runtime_state')

    def increase_backend_quotas_usage(self, validate=True):
        self.tenant.add_quota_usage(self.tenant.Quotas.floating_ip_count, 1, validate=validate)

    def decrease_backend_quotas_usage(self):
        self.tenant.add_quota_usage(self.tenant.Quotas.floating_ip_count, -1)


class Tenant(structure_models.PrivateCloud):

    class Quotas(QuotaModelMixin.Quotas):
        vcpu = QuotaField(default_limit=20, is_backend=True)
        ram = QuotaField(default_limit=51200, is_backend=True)
        storage = QuotaField(default_limit=1024000, is_backend=True)
        instances = QuotaField(default_limit=30, is_backend=True)
        security_group_count = QuotaField(default_limit=100, is_backend=True)
        security_group_rule_count = QuotaField(default_limit=100, is_backend=True)
        floating_ip_count = QuotaField(default_limit=50, is_backend=True)
        volumes = QuotaField(default_limit=50, is_backend=True)
        volumes_size = QuotaField(is_backend=True)
        snapshots = QuotaField(default_limit=50, is_backend=True)
        snapshots_size = QuotaField(is_backend=True)
        network_count = QuotaField(default_limit=10, is_backend=True)
        subnet_count = QuotaField(default_limit=10, is_backend=True)

    # backend_id is nullable on purpose, otherwise
    # it wouldn't be possible to put a unique constraint on it
    backend_id = models.CharField(max_length=255, blank=True, null=True)

    service_project_link = models.ForeignKey(
        OpenStackServiceProjectLink, related_name='tenants', on_delete=models.PROTECT)

    internal_network_id = models.CharField(max_length=64, blank=True)
    external_network_id = models.CharField(max_length=64, blank=True)
    availability_zone = models.CharField(
        max_length=100, blank=True,
        help_text=_('Optional availability group. Will be used for all instances provisioned in this tenant')
    )
    user_username = models.CharField(max_length=50, blank=True)
    user_password = models.CharField(max_length=50, blank=True)

    tracker = FieldTracker()

    class Meta(object):
        unique_together = ('service_project_link', 'backend_id')

    @classmethod
    def generate_username(cls, name):
        """
        Generates random valid tenant user name based on tenant name
        :param name: tenant name
        :return: username
        """
        return slugify(name)[:25] + '-user-%s' % core_utils.pwgen(4)

    def get_backend(self):
        return self.service_project_link.service.get_backend(tenant_id=self.backend_id)

    def get_log_fields(self):
        return super(Tenant, self).get_log_fields() + ('extra_configuration',)

    @classmethod
    def get_backend_fields(cls):
        return super(Tenant, cls).get_backend_fields() + ('name', 'description', 'error_message', 'runtime_state')

    def get_access_url(self):
        settings = self.service_project_link.service.settings
        access_url = settings.get_option('access_url')
        if access_url:
            return access_url

        if settings.backend_url:
            parsed = urllib.parse.urlparse(settings.backend_url)
            return '%s://%s/dashboard' % (parsed.scheme, parsed.hostname)

    def format_quota(self, name, limit):
        if name == self.Quotas.vcpu.name:
            return int(limit)
        elif name in (self.Quotas.storage.name, self.Quotas.ram.name):
            return _('%s GB') % int(limit / 1024)
        else:
            return limit


class Network(core_models.RuntimeStateMixin, structure_models.SubResource):
    service_project_link = models.ForeignKey(
        OpenStackServiceProjectLink, related_name='networks', on_delete=models.PROTECT)
    tenant = models.ForeignKey(Tenant, related_name='networks')
    is_external = models.BooleanField(default=False)
    type = models.CharField(max_length=50, blank=True)
    segmentation_id = models.IntegerField(null=True)

    def get_backend(self):
        return self.tenant.get_backend()

    @classmethod
    def get_url_name(cls):
        return 'openstack-network'

    def increase_backend_quotas_usage(self, validate=True):
        self.tenant.add_quota_usage(self.tenant.Quotas.network_count, 1, validate=validate)

    def decrease_backend_quotas_usage(self):
        self.tenant.add_quota_usage(self.tenant.Quotas.network_count, -1)

    @classmethod
    def get_backend_fields(cls):
        return super(Network, cls).get_backend_fields() + ('name', 'description', 'is_external', 'type',
                                                           'segmentation_id', 'runtime_state')


class SubNet(structure_models.SubResource):
    service_project_link = models.ForeignKey(
        OpenStackServiceProjectLink, related_name='subnets', on_delete=models.PROTECT)
    network = models.ForeignKey(Network, related_name='subnets')
    cidr = models.CharField(max_length=32, blank=True)
    gateway_ip = models.GenericIPAddressField(protocol='IPv4', null=True)
    allocation_pools = JSONField(default=dict)
    ip_version = models.SmallIntegerField(default=4)
    enable_dhcp = models.BooleanField(default=True)
    dns_nameservers = JSONField(default=list, help_text=_('List of DNS name servers associated with the subnet.'))

    class Meta:
        verbose_name = _('Subnet')
        verbose_name_plural = _('Subnets')

    def get_backend(self):
        return self.network.get_backend()

    @classmethod
    def get_url_name(cls):
        return 'openstack-subnet'

    def increase_backend_quotas_usage(self, validate=True):
        self.network.tenant.add_quota_usage(self.network.tenant.Quotas.subnet_count, 1, validate=validate)

    def decrease_backend_quotas_usage(self):
        self.network.tenant.add_quota_usage(self.network.tenant.Quotas.subnet_count, -1)

    @classmethod
    def get_backend_fields(cls):
        return super(SubNet, cls).get_backend_fields() + ('name', 'description', 'allocation_pools', 'cidr',
                                                          'ip_version', 'enable_dhcp', 'gateway_ip', 'dns_nameservers')


class CustomerOpenStack(TimeStampedModel):
    settings = models.ForeignKey(structure_models.ServiceSettings,
                                 on_delete=models.CASCADE,
                                 limit_choices_to={'shared': True, 'type': 'OpenStack'}, )
    customer = models.ForeignKey(structure_models.Customer, on_delete=models.CASCADE)
    external_network_id = models.CharField(_('OpenStack external network ID'), max_length=255)

    class Meta(object):
        verbose_name = _('Organization OpenStack settings')
        verbose_name_plural = _('Organization OpenStack settings')
        unique_together = ('settings', 'customer')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2018-03-02 08:20
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0012_backend_id_null'),
    ]

    operations = [
        migrations.AddField(
            model_name='issue',
            name='assignee',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='issue',
            name='creator_name',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='issue',
            name='resolution_date',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging

from django.db import models, migrations
import waldur_core.structure.models

logger = logging.getLogger(__name__)


def create_service_type(apps, schema_editor):
    service_types = {
        1: 'OpenStack',
        2: 'DigitalOcean',
        3: 'Amazon',
        4: 'Jira',
        5: 'GitLab',
        6: 'Oracle',
        7: 'Azure',
        8: 'SugarCRM',
        9: 'SaltStack',
        10: 'Zabbix'
    }

    ServiceSettings = apps.get_model('structure', 'ServiceSettings')
    for service in ServiceSettings.objects.all():
        try:
            service.service_type = service_types[service.type]
            service.save()
        except KeyError:
            if service.type in service_types.values():
                service.service_type = service.type
                service.save()
            else:
                logger.warning('Cannot migrate service type %s' % service.type)


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0027_servicesettings_service_type'),
    ]

    operations = [
        migrations.RunPython(create_service_type),
        migrations.RemoveField(
            model_name='servicesettings',
            name='type',
        ),
        migrations.RenameField(
            model_name='servicesettings',
            old_name='service_type',
            new_name='type'
        ),
        migrations.AlterField(
            model_name='servicesettings',
            name='type',
            field=models.CharField(max_length=255, db_index=True, validators=[waldur_core.structure.models.validate_service_type]),
            preserve_default=True,
        )
    ]



from django.contrib.contenttypes.models import ContentType

from waldur_core.cost_tracking import ConsumableItem
from waldur_core.cost_tracking.models import DefaultPriceListItem

from . import models, PriceItemTypes


def get_consumable_item(flavor_name):
    return ConsumableItem(item_type=PriceItemTypes.FLAVOR, key=flavor_name, name='Flavor: %s' % flavor_name)


def sync_price_list_item(flavor):
    resource_content_type = ContentType.objects.get_for_model(models.Instance)
    consumable_item = get_consumable_item(flavor.name)
    DefaultPriceListItem._create_or_update_default_price_list_item(
        resource_content_type=resource_content_type,
        consumable_item=consumable_item,
    )



from django.core.management.base import BaseCommand
from waldur_mastermind.marketplace_packages import utils


class Command(BaseCommand):
    help = 'Init marketplace offerings for OpenStack package templates.'

    def handle(self, *args, **options):
        utils.create_missing_offerings()



from django.conf import settings
from waldur_ansible.common import cache_utils
from waldur_ansible.python_management import models

PIP_SYNCING_LOCK = 'waldur_syncing_pip_packages'

PYTHON_MANAGEMENT_GLOBAL_LOCK = 'waldur_python_management_global_%s'
PYTHON_MANAGEMENT_VIRTUAL_ENV_SYNCING_LOCK = 'waldur_python_management_%s_%s'


class NullProcessingFinishedLockingHandler(object):
    def handle_on_processing_finished(self, request):
        pass


class RelatedToVirtualEnvRequestProcessingFinishedLockingHandler(object):
    def handle_on_processing_finished(self, request):
        cache_utils.release_task_status(
            PythonManagementBackendLockBuilder.build_related_to_virt_env_lock(
                request.python_management, request.virtual_env_name))


class GlobalRequestProcessingFinishedLockingHandler(object):
    def handle_on_processing_finished(self, request):
        cache_utils.release_task_status(
            PythonManagementBackendLockBuilder.build_global_lock(request.python_management))


class NullProcessingAllowedDecider(object):
    def is_processing_allowed(self, request):
        return True


class RelatedToVirtualEnvProcessingAllowedDecider(object):
    def is_processing_allowed(self, request):
        virtual_env_lock = PythonManagementBackendLockBuilder.build_related_to_virt_env_lock(
            request.python_management, request.virtual_env_name)
        global_lock = PythonManagementBackendLockBuilder.build_global_lock(request.python_management)
        return not cache_utils.is_syncing(virtual_env_lock) and not cache_utils.is_syncing(global_lock)


class GlobalProcessingAllowedDecider(object):
    def is_processing_allowed(self, request):
        global_lock = PythonManagementBackendLockBuilder.build_global_lock(request.python_management)
        return not cache_utils.is_syncing(global_lock)


class NullSynchronizer(object):
    def lock(self, request):
        pass


class RelatedToVirtualEnvSynchronizer(object):
    def lock(self, request):
        virtual_env_lock = PythonManagementBackendLockBuilder.build_related_to_virt_env_lock(
            request.python_management, request.virtual_env_name)
        cache_utils.renew_task_status(virtual_env_lock, settings.WALDUR_ANSIBLE_COMMON['ANSIBLE_REQUEST_TIMEOUT'])


class GlobalSynchronizer(object):
    def lock(self, request):
        global_lock = PythonManagementBackendLockBuilder.build_global_lock(request.python_management)
        cache_utils.renew_task_status(global_lock, settings.WALDUR_ANSIBLE_COMMON['ANSIBLE_REQUEST_TIMEOUT'])


class PythonManagementBackendLockingService(object):

    @staticmethod
    def lock_for_processing(request):
        synchronizer = PythonManagementBackendLockBuilder.intantiate_synchronizer(type(request))
        return synchronizer.lock(request)

    @staticmethod
    def is_processing_allowed(request):
        processing_allowed_decider = PythonManagementBackendLockBuilder.intantiate_processing_allowed_decider(type(request))
        return processing_allowed_decider.is_processing_allowed(request)

    @staticmethod
    def handle_on_processing_finished(request):
        locking_handler = PythonManagementBackendLockBuilder.intantiate_processing_finished_locking_handler(type(request))
        locking_handler.handle_on_processing_finished(request)


class PythonManagementBackendLockBuilder(object):
    # Initialize requests do not need locking - there can be no duplicates of same PythonManagement model
    REQUEST_TYPES_PROCESSING_ALLOWED_DECIDER = {
        models.PythonManagementInitializeRequest: NullProcessingAllowedDecider,
        models.PythonManagementSynchronizeRequest: RelatedToVirtualEnvProcessingAllowedDecider,
        models.PythonManagementFindVirtualEnvsRequest: GlobalProcessingAllowedDecider,
        models.PythonManagementFindInstalledLibrariesRequest: RelatedToVirtualEnvProcessingAllowedDecider,
        models.PythonManagementDeleteVirtualEnvRequest: RelatedToVirtualEnvProcessingAllowedDecider,
        models.PythonManagementDeleteRequest: GlobalProcessingAllowedDecider,
    }
    REQUEST_TYPES_SYNCHRONIZERS = {
        models.PythonManagementInitializeRequest: NullSynchronizer,
        models.PythonManagementSynchronizeRequest: RelatedToVirtualEnvSynchronizer,
        models.PythonManagementFindVirtualEnvsRequest: GlobalSynchronizer,
        models.PythonManagementFindInstalledLibrariesRequest: RelatedToVirtualEnvSynchronizer,
        models.PythonManagementDeleteVirtualEnvRequest: RelatedToVirtualEnvSynchronizer,
        models.PythonManagementDeleteRequest: GlobalSynchronizer,
    }
    REQUEST_TYPES_PROCESSING_FINISHED_LOCKING_HANDLER = {
        models.PythonManagementInitializeRequest: NullProcessingFinishedLockingHandler,
        models.PythonManagementSynchronizeRequest: RelatedToVirtualEnvRequestProcessingFinishedLockingHandler,
        models.PythonManagementFindVirtualEnvsRequest: GlobalRequestProcessingFinishedLockingHandler,
        models.PythonManagementFindInstalledLibrariesRequest: RelatedToVirtualEnvRequestProcessingFinishedLockingHandler,
        models.PythonManagementDeleteVirtualEnvRequest: RelatedToVirtualEnvRequestProcessingFinishedLockingHandler,
        models.PythonManagementDeleteRequest: GlobalRequestProcessingFinishedLockingHandler,
    }

    @staticmethod
    def intantiate_processing_allowed_decider(python_management_request_class):
        processing_allowed_decider_class = PythonManagementBackendLockBuilder.REQUEST_TYPES_PROCESSING_ALLOWED_DECIDER \
            .get(python_management_request_class)
        return processing_allowed_decider_class()

    @staticmethod
    def intantiate_processing_finished_locking_handler(python_management_request_class):
        locking_handler_class = PythonManagementBackendLockBuilder.REQUEST_TYPES_PROCESSING_FINISHED_LOCKING_HANDLER \
            .get(python_management_request_class)
        return locking_handler_class()

    @staticmethod
    def intantiate_synchronizer(python_management_request_class):
        locking_handler_class = PythonManagementBackendLockBuilder.REQUEST_TYPES_SYNCHRONIZERS \
            .get(python_management_request_class)
        return locking_handler_class()

    @staticmethod
    def build_related_to_virt_env_lock(persisted_python_management, virtual_env_name):
        return PYTHON_MANAGEMENT_VIRTUAL_ENV_SYNCING_LOCK % (persisted_python_management.pk, virtual_env_name)

    @staticmethod
    def build_global_lock(persisted_python_management):
        return PYTHON_MANAGEMENT_GLOBAL_LOCK % persisted_python_management.pk



from __future__ import unicode_literals

import logging
import re

from django.conf import settings
from django.template.loader import render_to_string
from django.utils import timezone

from waldur_core.core import utils
from waldur_core.core.models import StateMixin
from waldur_core.core.tasks import send_task
from waldur_core.structure import SupportedServices, signals
from waldur_core.structure.log import event_logger
from waldur_core.structure.models import (Customer, CustomerPermission, Project, ProjectPermission,
                                          Service, ServiceSettings, CustomerRole)

logger = logging.getLogger(__name__)


def revoke_roles_on_project_deletion(sender, instance=None, **kwargs):
    """
    When project is deleted, all project permissions are cascade deleted
    by Django without emitting structure_role_revoked signal.
    So in order to invalidate nc_user_count quota we need to emit it manually.
    """
    instance.remove_all_users()


def log_customer_save(sender, instance, created=False, **kwargs):
    if created:
        event_logger.customer.info(
            'Customer {customer_name} has been created.',
            event_type='customer_creation_succeeded',
            event_context={
                'customer': instance,
            })
    else:
        event_logger.customer.info(
            'Customer {customer_name} has been updated.',
            event_type='customer_update_succeeded',
            event_context={
                'customer': instance,
            })


def log_customer_delete(sender, instance, **kwargs):
    event_logger.customer.info(
        'Customer {customer_name} has been deleted.',
        event_type='customer_deletion_succeeded',
        event_context={
            'customer': instance,
        })


def log_project_save(sender, instance, created=False, **kwargs):
    if created:
        event_logger.project.info(
            'Project {project_name} has been created.',
            event_type='project_creation_succeeded',
            event_context={
                'project': instance,
            })
    else:
        changed_fields = instance.tracker.changed().copy()
        changed_fields.pop('modified', None)
        if not changed_fields:
            return

        message = 'Project {project_name} has been updated.'
        for name, previous_value in changed_fields.items():
            current_value = getattr(instance, name)
            message = "%s %s has been changed from '%s' to '%s'." % (
                message,
                name.capitalize(),
                previous_value,
                current_value,
            )

        event_logger.project.info(message, event_type='project_update_succeeded', event_context={'project': instance})


def log_project_delete(sender, instance, **kwargs):
    event_logger.project.info(
        'Project {project_name} has been deleted.',
        event_type='project_deletion_succeeded',
        event_context={
            'project': instance,
        })


def log_customer_role_granted(sender, structure, user, role, created_by=None, **kwargs):
    event_context = {
        'customer': structure,
        'affected_user': user,
        'structure_type': 'customer',
        'role_name': CustomerPermission(role=role).get_role_display(),
    }
    if created_by:
        event_context['user'] = created_by

    event_logger.customer_role.info(
        'User {affected_user_username} has gained role of {role_name} in customer {customer_name}.',
        event_type='role_granted', event_context=event_context)


def log_customer_role_revoked(sender, structure, user, role, removed_by=None, **kwargs):
    event_context = {
        'customer': structure,
        'affected_user': user,
        'structure_type': 'customer',
        'role_name': CustomerPermission(role=role).get_role_display(),
    }
    if removed_by:
        event_context['user'] = removed_by

    event_logger.customer_role.info(
        'User {affected_user_username} has lost role of {role_name} in customer {customer_name}.',
        event_type='role_revoked', event_context=event_context)


def log_customer_role_updated(sender, instance, user, **kwargs):
    template = 'User %(user_username)s has changed permission expiration time ' \
               'for user {affected_user_username} in customer {customer_name} from ' \
               '%(old_expiration_time)s to %(new_expiration_time)s.'

    context = {
        'old_expiration_time': instance.tracker.previous('expiration_time'),
        'new_expiration_time': instance.expiration_time,
        'user_username': user.full_name or user.username,
    }

    event_logger.customer_role.info(
        template % context,
        event_type='role_updated',
        event_context={
            'customer': instance.customer,
            'affected_user': instance.user,
            'structure_type': 'customer',
            'role_name': instance.get_role_display(),
        })


def log_project_role_granted(sender, structure, user, role, created_by=None, **kwargs):
    event_context = {
        'project': structure,
        'affected_user': user,
        'structure_type': 'project',
        'role_name': ProjectPermission(role=role).get_role_display(),
    }
    if created_by:
        event_context['user'] = created_by

    event_logger.project_role.info(
        'User {affected_user_username} has gained role of {role_name} in project {project_name}.',
        event_type='role_granted', event_context=event_context)


def log_project_role_revoked(sender, structure, user, role, removed_by=None, **kwargs):
    event_context = {
        'project': structure,
        'affected_user': user,
        'structure_type': 'project',
        'role_name': ProjectPermission(role=role).get_role_display()
    }
    if removed_by:
        event_context['user'] = removed_by

    event_logger.project_role.info(
        'User {affected_user_username} has revoked role of {role_name} in project {project_name}.',
        event_type='role_revoked', event_context=event_context)


def log_project_role_updated(sender, instance, user, **kwargs):
    template = 'User %(user_username)s has changed permission expiration time ' \
               'for user {affected_user_username} in project {project_name} from ' \
               '%(old_expiration_time)s to %(new_expiration_time)s.'

    context = {
        'old_expiration_time': instance.tracker.previous('expiration_time'),
        'new_expiration_time': instance.expiration_time,
        'user_username': user.full_name or user.username,
    }

    event_logger.project_role.info(
        template % context,
        event_type='role_updated',
        event_context={
            'project': instance.project,
            'affected_user': instance.user,
            'structure_type': 'project',
            'role_name': instance.get_role_display(),
        })


def change_customer_nc_users_quota(sender, structure, user, role, signal, **kwargs):
    """ Modify nc_user_count quota usage on structure role grant or revoke """
    assert signal in (signals.structure_role_granted, signals.structure_role_revoked), \
        'Handler "change_customer_nc_users_quota" has to be used only with structure_role signals'
    assert sender in (Customer, Project), \
        'Handler "change_customer_nc_users_quota" works only with Project and Customer models'

    if sender == Customer:
        customer = structure
    elif sender == Project:
        customer = structure.customer

    customer_users = customer.get_users()
    customer.set_quota_usage(Customer.Quotas.nc_user_count, customer_users.count())


def log_resource_deleted(sender, instance, **kwargs):
    event_logger.resource.info(
        '{resource_full_name} has been deleted.',
        event_type='resource_deletion_succeeded',
        event_context={'resource': instance})


def log_resource_imported(sender, instance, **kwargs):
    event_logger.resource.info(
        'Resource {resource_full_name} has been imported.',
        event_type='resource_import_succeeded',
        event_context={'resource': instance})


def log_resource_creation_succeeded(instance):
    event_logger.resource.info(
        'Resource {resource_name} has been created.',
        event_type='resource_creation_succeeded',
        event_context={'resource': instance})


def log_resource_creation_failed(instance):
    event_logger.resource.error(
        'Resource {resource_name} creation has failed.',
        event_type='resource_creation_failed',
        event_context={'resource': instance})


def log_resource_creation_scheduled(sender, instance, created=False, **kwargs):
    if created and isinstance(instance, StateMixin) and instance.state == StateMixin.States.CREATION_SCHEDULED:
        event_logger.resource.info(
            'Resource {resource_name} creation has been scheduled.',
            event_type='resource_creation_scheduled',
            event_context={'resource': instance},
        )


def log_resource_action(sender, instance, name, source, target, **kwargs):
    if isinstance(instance, StateMixin):
        if source == StateMixin.States.CREATING:
            if target == StateMixin.States.OK:
                log_resource_creation_succeeded(instance)
            elif target == StateMixin.States.ERRED:
                log_resource_creation_failed(instance)

    if isinstance(instance, StateMixin) and target == StateMixin.States.DELETION_SCHEDULED:
        event_logger.resource.info(
            'Resource {resource_name} deletion has been scheduled.',
            event_type='resource_deletion_scheduled',
            event_context={'resource': instance},
        )


def detect_vm_coordinates(sender, instance, name, source, target, **kwargs):
    # Check if geolocation is enabled
    if not settings.WALDUR_CORE.get('ENABLE_GEOIP', True):
        return

    # VM already has coordinates
    if instance.latitude is not None and instance.longitude is not None:
        return

    if target == StateMixin.States.OK:
        send_task('structure', 'detect_vm_coordinates')(utils.serialize_instance(instance))


def connect_customer_to_shared_service_settings(sender, instance, created=False, **kwargs):
    if not created:
        return
    customer = instance

    for shared_settings in ServiceSettings.objects.filter(shared=True):
        try:
            service_model = SupportedServices.get_service_models()[shared_settings.type]['service']
            service_model.objects.create(customer=customer,
                                         settings=shared_settings,
                                         available_for_all=True)
        except KeyError:
            logger.warning("Unregistered service of type %s" % shared_settings.type)


def connect_project_to_all_available_services(sender, instance, created=False, **kwargs):
    if not created:
        return
    project = instance

    for service_model in Service.get_all_models():
        for service in service_model.objects.filter(available_for_all=True, customer=project.customer):
            service_project_link_model = service.projects.through
            service_project_link_model.objects.create(project=project, service=service)


def connect_service_to_all_projects_if_it_is_available_for_all(sender, instance, created=False, **kwargs):
    service = instance
    if service.available_for_all:
        service_project_link_model = service.projects.through
        for project in service.customer.projects.all():
            service_project_link_model.objects.get_or_create(project=project, service=service)


def delete_service_settings_on_service_delete(sender, instance, **kwargs):
    """ Delete not shared service settings without services """
    service = instance
    try:
        service_settings = service.settings
    except ServiceSettings.DoesNotExist:
        # If this handler works together with delete_service_settings_on_scope_delete
        # it tries to delete service settings that are already deleted.
        return
    if not service_settings.shared:
        service.settings.delete()


def update_resource_start_time(sender, instance, created=False, **kwargs):
    if created:
        return

    if not instance.tracker.has_changed('runtime_state'):
        return

    # queryset is needed in order to call update method which does not
    # emit post_save signal, otherwise it's called recursively
    queryset = instance._meta.model.objects.filter(pk=instance.pk)

    if instance.runtime_state == instance.get_online_state():
        queryset.update(start_time=timezone.now())

    if instance.runtime_state == instance.get_offline_state():
        queryset.update(start_time=None)


def delete_service_settings_on_scope_delete(sender, instance, **kwargs):
    """ If VM that contains service settings were deleted - all settings
        resources could be safely deleted from NC.
    """
    for service_settings in ServiceSettings.objects.filter(scope=instance):
        service_settings.unlink_descendants()
        service_settings.delete()


def clean_tags_cache_after_tagged_item_saved(sender, instance, **kwargs):
    instance.content_object.clean_tag_cache()


def clean_tags_cache_before_tagged_item_deleted(sender, instance, **kwargs):
    instance.content_object.clean_tag_cache()


def notify_about_user_profile_changes(sender, instance, created=False, **kwargs):
    if created or not settings.WALDUR_CORE['NOTIFICATIONS_PROFILE_CHANGES']['ENABLED']:
        return

    user = instance
    change_fields = settings.WALDUR_CORE['NOTIFICATIONS_PROFILE_CHANGES']['FIELDS']
    organizations = Customer.objects.filter(
        permissions__user=user,
        permissions__is_active=True,
        permissions__role=CustomerRole.OWNER)

    if not ((set(change_fields) & set(user.tracker.changed())) and organizations):
        return

    fields = []
    for field in change_fields:
        if user.tracker.has_changed(field):
            fields.append({
                'name': field,
                'old_value': user.tracker.previous(field),
                'new_value': getattr(user, field, None)
            })

    msg = render_to_string('structure/notifications_profile_changes.html', {
        'user': user,
        'fields': fields,
        'organizations': organizations
    })

    msg = re.sub(r'\s+', ' ', msg).strip()

    event_logger.user.info(
        msg,
        event_type='user_profile_changed',
        event_context={'affected_user': user})



from django.db import transaction

from . import utils


def create_offering_for_package_template(sender, instance, created=False, **kwargs):
    if created:
        utils.create_offering_for_package_template(instance)
    else:
        utils.update_offering_for_template(instance)


def sync_offering_attribute_with_template_component(sender, instance, created=False, **kwargs):
    transaction.on_commit(lambda: utils.sync_offering_attribute_with_template_component(instance))



from __future__ import unicode_literals

from django.core.exceptions import PermissionDenied
from django.db.models import Count
from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import response, viewsets, permissions, status, decorators, mixins

from waldur_core.core import serializers as core_serializers, filters as core_filters, permissions as core_permissions
from waldur_core.core.managers import SummaryQuerySet
from waldur_core.logging import elasticsearch_client, models, serializers, filters, utils
from waldur_core.logging.loggers import get_event_groups, get_alert_groups, event_logger


class EventViewSet(mixins.CreateModelMixin, viewsets.GenericViewSet):
    permission_classes = (permissions.IsAuthenticated, core_permissions.IsAdminOrReadOnly)
    filter_backends = (filters.EventFilterBackend,)
    serializer_class = serializers.EventSerializer

    def get_queryset(self):
        return elasticsearch_client.ElasticsearchResultList()

    def list(self, request, *args, **kwargs):
        """
        To get a list of events - run **GET** against */api/events/* as authenticated user. Note that a user can
        only see events connected to objects she is allowed to see.

        Sorting is supported in ascending and descending order by specifying a field to an **?o=** parameter. By default
        events are sorted by @timestamp in descending order.

        Run POST against */api/events/* to create an event. Only users with staff privileges can create events.
        New event will be emitted with `custom_notification` event type.
        Request should contain following fields:

        - level: the level of current event. Following levels are supported: debug, info, warning, error
        - message: string representation of event message
        - scope: optional URL, which points to the loggable instance

        Request example:

        .. code-block:: javascript

            POST /api/events/
            Accept: application/json
            Content-Type: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "level": "info",
                "message": "message#1",
                "scope": "http://example.com/api/customers/9cd869201e1b4158a285427fcd790c1c/"
            }
        """
        self.queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(self.queryset)
        if page is not None:
            return self.get_paginated_response(page)
        return response.Response(self.queryset)

    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def perform_create(self, serializer):
        scope = serializer.validated_data.get('scope')
        context = {'scope': scope} if scope is not None else {}

        event_logger.custom.process(
            level=serializer.validated_data.get('level'),
            message_template=serializer.validated_data.get('message'),
            event_type='custom_notification',
            event_context=context,
            fail_silently=False
        )

    @decorators.list_route()
    def count(self, request, *args, **kwargs):
        """
        To get a count of events - run **GET** against */api/events/count/* as authenticated user.
        Endpoint support same filters as events list.

        Response example:

        .. code-block:: javascript

            {"count": 12321}
        """

        self.queryset = self.filter_queryset(self.get_queryset())
        return response.Response({'count': self.queryset.count()}, status=status.HTTP_200_OK)

    @decorators.list_route()
    def count_history(self, request, *args, **kwargs):
        """
        To get a historical data of events amount - run **GET** against */api/events/count/history/*.
        Endpoint support same filters as events list.
        More about historical data - read at section *Historical data*.

        Response example:

        .. code-block:: javascript

            [
                {
                    "point": 141111111111,
                    "object": {
                        "count": 558
                    }
                }
            ]
        """
        queryset = self.filter_queryset(self.get_queryset())

        mapped = {
            'start': request.query_params.get('start'),
            'end': request.query_params.get('end'),
            'points_count': request.query_params.get('points_count'),
            'point_list': request.query_params.getlist('point'),
        }
        serializer = core_serializers.HistorySerializer(data={k: v for k, v in mapped.items() if v})
        serializer.is_valid(raise_exception=True)

        timestamp_ranges = [{'end': point_date} for point_date in serializer.get_filter_data()]
        aggregated_count = queryset.aggregated_count(timestamp_ranges)

        return response.Response(
            [{'point': int(ac['end']), 'object': {'count': ac['count']}} for ac in aggregated_count],
            status=status.HTTP_200_OK)

    @decorators.list_route()
    def scope_types(self, request, *args, **kwargs):
        """ Returns a list of scope types acceptable by events filter. """
        return response.Response(utils.get_scope_types_mapping().keys())

    @decorators.list_route()
    def event_groups(self, request, *args, **kwargs):
        """
        Returns a list of groups with event types.
        Group is used in exclude_features query param.
        """
        return response.Response(get_event_groups())


class AlertViewSet(mixins.CreateModelMixin,
                   viewsets.ReadOnlyModelViewSet):
    queryset = models.Alert.objects.all()
    serializer_class = serializers.AlertSerializer
    lookup_field = 'uuid'
    filter_backends = (
        DjangoFilterBackend,
        filters.AdditionalAlertFilterBackend,
        filters.ExternalAlertFilterBackend,
        filters.AlertScopeFilterBackend,
    )
    filter_class = filters.AlertFilter

    def get_queryset(self):
        return models.Alert.objects.filtered_for_user(self.request.user).order_by('-created')

    def list(self, request, *args, **kwargs):
        """
        To get a list of alerts, run **GET** against */api/alerts/* as authenticated user.

        Alert severity field can take one of this values: "Error", "Warning", "Info", "Debug".
        Field scope will contain link to object that cause alert.
        Context - dictionary that contains information about all related to alert objects.

        Alerts can be filtered by:
         - ?severity=<severity> (can be list)
         - ?alert_type=<alert_type> (can be list)
         - ?scope=<url> concrete alert scope
         - ?scope_type=<string> name of scope type (Ex.: instance, service_project_link, project...)
           DEPRECATED use ?content_type instead
         - ?created_from=<timestamp>
         - ?created_to=<timestamp>
         - ?closed_from=<timestamp>
         - ?closed_to=<timestamp>
         - ?from=<timestamp> - filter alerts that was active from given date
         - ?to=<timestamp> - filter alerts that was active to given date
         - ?opened - if this argument is in GET request endpoint will return only alerts that are not closed
         - ?closed - if this argument is in GET request endpoint will return only alerts that are closed
         - ?aggregate=aggregate_model_name (default: 'customer'. Have to be from list: 'customer', project')
         - ?uuid=uuid_of_aggregate_model_object (not required. If this parameter will be defined - result ill contain only
           object with given uuid)
         - ?acknowledged=True|False - show only acknowledged (non-acknowledged) alerts
         - ?content_type=<string> name of scope content type in format <app_name>.<scope_type>
           (Ex.: structure.project, openstack.instance...)
         - ?exclude_features=<feature> (can be list) - exclude alert from output if it's type corresponds o one of given features

        Alerts can be ordered by:

         -?o=severity - order by severity
         -?o=created - order by creation time

        .. code-block:: http

            GET /api/alerts/
            Accept: application/json
            Content-Type: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            [
                {
                    "url": "http://example.com/api/alerts/e80e48a4e58b48ff9a1320a0aa0d68ab/",
                    "uuid": "e80e48a4e58b48ff9a1320a0aa0d68ab",
                    "alert_type": "first_alert",
                    "message": "message#1",
                    "severity": "Debug",
                    "scope": "http://example.com/api/instances/9d1d7e03b0d14fd0b42b5f649dfa3de5/",
                    "created": "2015-05-29T14:24:27.342Z",
                    "closed": null,
                    "context": {
                        'customer_abbreviation': 'customer_abbreviation',
                        'customer_contact_details': 'customer details',
                        'customer_name': 'Customer name',
                        'customer_uuid': '53c6e86406e349faa7924f4c865b15ab',
                        'quota_limit': '131072.0',
                        'quota_name': 'ram',
                        'quota_usage': '131071',
                        'quota_uuid': 'f6ae2f7ca86f4e2f9bb64de1015a2815',
                        'scope_name': 'project X',
                        'scope_uuid': '0238d71ee1934bd2839d4e71e5f9b91a'
                    }
                    "acknowledged": true,
                }
            ]
        """
        return super(AlertViewSet, self).list(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        """
        Run **POST** against */api/alerts/* to create or update alert. If alert with posted scope and
        alert_type already exists - it will be updated. Only users with staff privileges can create alerts.

        Request example:

        .. code-block:: javascript

            POST /api/alerts/
            Accept: application/json
            Content-Type: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "scope": "http://testserver/api/projects/b9e8a102b5ff4469b9ac03253fae4b95/",
                "message": "message#1",
                "alert_type": "first_alert",
                "severity": "Debug"
            }
        """
        return super(AlertViewSet, self).create(request, *args, **kwargs)

    @decorators.detail_route(methods=['post'])
    def close(self, request, *args, **kwargs):
        """
        To close alert - run **POST** against */api/alerts/<alert_uuid>/close/*. No data is required.
        Only users with staff privileges can close alerts.
        """
        if not request.user.is_staff:
            raise PermissionDenied()
        alert = self.get_object()
        alert.close()

        return response.Response(status=status.HTTP_204_NO_CONTENT)

    @decorators.detail_route(methods=['post'])
    def acknowledge(self, request, *args, **kwargs):
        """
        To acknowledge alert - run **POST** against */api/alerts/<alert_uuid>/acknowledge/*. No payload is required.
        All users that can see alerts can also acknowledge it. If alert is already acknowledged endpoint
        will return error with code 409(conflict).
        """
        alert = self.get_object()
        if not alert.acknowledged:
            alert.acknowledge()
            return response.Response(status=status.HTTP_200_OK)
        else:
            return response.Response({'detail': _('Alert is already acknowledged.')}, status=status.HTTP_409_CONFLICT)

    @decorators.detail_route(methods=['post'])
    def cancel_acknowledgment(self, request, *args, **kwargs):
        """
        To cancel alert acknowledgment - run **POST** against */api/alerts/<alert_uuid>/cancel_acknowledgment/*.
        No payload is required. All users that can see alerts can also cancel it acknowledgment.
        If alert is not acknowledged endpoint will return error with code 409 (conflict).
        """
        alert = self.get_object()
        if alert.acknowledged:
            alert.cancel_acknowledgment()
            return response.Response(status=status.HTTP_200_OK)
        else:
            return response.Response({'detail': _('Alert is not acknowledged.')}, status=status.HTTP_409_CONFLICT)

    @decorators.list_route()
    def stats(self, request, *args, **kwargs):
        """
        To get count of alerts per severities - run **GET** request against */api/alerts/stats/*.
        This endpoint supports all filters that are available for alerts list (*/api/alerts/*).

        Response example:

        .. code-block:: javascript

            {
                "debug": 2,
                "error": 1,
                "info": 1,
                "warning": 1
            }
        """
        queryset = self.filter_queryset(self.get_queryset())
        alerts_severities_count = queryset.values('severity').annotate(count=Count('severity'))

        severity_names = dict(models.Alert.SeverityChoices.CHOICES)
        # For consistency with all other endpoint we need to return severity names in lower case.
        alerts_severities_count = {
            severity_names[asc['severity']].lower(): asc['count'] for asc in alerts_severities_count}
        for severity_name in severity_names.values():
            if severity_name.lower() not in alerts_severities_count:
                alerts_severities_count[severity_name.lower()] = 0

        return response.Response(alerts_severities_count, status=status.HTTP_200_OK)

    def perform_create(self, serializer):
        if not self.request.user.is_staff:
            raise PermissionDenied()

        super(AlertViewSet, self).perform_create(serializer)

    @decorators.list_route()
    def alert_groups(self, request, *args, **kwargs):
        """
        Returns a list of groups with alert types.
        Group is used in exclude_features query param.
        """
        return response.Response(get_alert_groups())


class BaseHookViewSet(viewsets.ModelViewSet):
    """
    Hooks API allows user to receive event notifications via different channel, like email or webhook.
    To get a list of all your hooks, run **GET** against */api/hooks/* as an authenticated user.
    """
    filter_backends = (core_filters.StaffOrUserFilter, DjangoFilterBackend)
    lookup_field = 'uuid'


class WebHookViewSet(BaseHookViewSet):
    queryset = models.WebHook.objects.all()
    filter_class = filters.WebHookFilter
    serializer_class = serializers.WebHookSerializer

    def create(self, request, *args, **kwargs):
        """
        To create new web hook issue **POST** against */api/hooks-web/* as an authenticated user.
        You should specify list of event_types or event_groups.

        Example of a request:

        .. code-block:: http

            POST /api/hooks-web/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "event_types": ["resource_start_succeeded"],
                "event_groups": ["users"],
                "destination_url": "http://example.com/"
            }

        When hook is activated, **POST** request is issued against destination URL with the following data:

        .. code-block:: javascript

            {
                "timestamp": "2015-07-14T12:12:56.000000",
                "message": "Customer ABC LLC has been updated.",
                "type": "customer_update_succeeded",
                "context": {
                    "user_native_name": "Walter Lebrowski",
                    "customer_contact_details": "",
                    "user_username": "Walter",
                    "user_uuid": "1c3323fc4ae44120b57ec40dea1be6e6",
                    "customer_uuid": "4633bbbb0b3a4b91bffc0e18f853de85",
                    "ip_address": "8.8.8.8",
                    "user_full_name": "Walter Lebrowski",
                    "customer_abbreviation": "ABC LLC",
                    "customer_name": "ABC LLC"
                },
                "levelname": "INFO"
            }

        Note that context depends on event type.
        """
        return super(WebHookViewSet, self).create(request, *args, **kwargs)


class EmailHookViewSet(BaseHookViewSet):
    queryset = models.EmailHook.objects.all()
    filter_class = filters.EmailHookFilter
    serializer_class = serializers.EmailHookSerializer

    def create(self, request, *args, **kwargs):
        """
        To create new email hook issue **POST** against */api/hooks-email/* as an authenticated user.
        You should specify list of event_types or event_groups.

        Example of a request:

        .. code-block:: http

            POST /api/hooks-email/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "event_types": ["openstack_instance_start_succeeded"],
                "event_groups": ["users"],
                "email": "test@example.com"
            }

        You may temporarily disable hook without deleting it by issuing following **PATCH** request against hook URL:

        .. code-block:: javascript

            {
                "is_active": "false"
            }
        """
        return super(EmailHookViewSet, self).create(request, *args, **kwargs)


class PushHookViewSet(BaseHookViewSet):
    queryset = models.PushHook.objects.all()
    filter_class = filters.PushHookFilter
    serializer_class = serializers.PushHookSerializer

    def create(self, request, *args, **kwargs):
        """
        To create new push hook issue **POST** against */api/hooks-push/* as an authenticated user.
        You should specify list of event_types or event_groups.

        Example of a request:

        .. code-block:: http

            POST /api/hooks-push/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "event_types": ["resource_start_succeeded"],
                "event_groups": ["users"],
                "type": "Android"
            }

        You may temporarily disable hook without deleting it by issuing following **PATCH** request against hook URL:

        .. code-block:: javascript

            {
                "is_active": "false"
            }
        """
        return super(PushHookViewSet, self).create(request, *args, **kwargs)


class HookSummary(mixins.ListModelMixin, viewsets.GenericViewSet):
    """
    Use */api/hooks/* to get a list of all the hooks of any type that a user can see.
    """
    serializer_class = serializers.SummaryHookSerializer
    filter_backends = (core_filters.StaffOrUserFilter, filters.HookSummaryFilterBackend)

    def get_queryset(self):
        return SummaryQuerySet(models.BaseHook.get_all_models())



from django.apps import AppConfig
from django.db.models import signals


class AzureConfig(AppConfig):
    name = 'waldur_azure'
    verbose_name = 'Waldur Azure'
    service_name = 'Azure'
    is_public_service = True

    def ready(self):
        from waldur_core.structure import SupportedServices
        from backend import AzureBackend
        SupportedServices.register_backend(AzureBackend)

        from waldur_azure import models, handlers

        signals.post_save.connect(
            handlers.copy_cloud_service_name_on_service_creation,
            sender=models.AzureServiceProjectLink,
            dispatch_uid='waldur_azure.handlers.copy_cloud_service_name_on_service_creation',
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-02-13 21:26
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0009_drop_old_fields'),
    ]

    operations = [
        migrations.AddField(
            model_name='issue',
            name='resolution_sla',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]
